# @file test_conf_mgmt.py
# Unit test suite for the ConfMgmt class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import tempfile
import os
import shutil
import logging
from edk2toolext.environment import conf_mgmt

test_file_text = '''#
#  Copyright (c) 2006 - 2018, Intel Corporation. All rights reserved.<BR>
#  Portions copyright (c) 2008 - 2009, Apple Inc. All rights reserved.<BR>
#  Portions copyright (c) 2011 - 2014, ARM Ltd. All rights reserved.<BR>
#  Copyright (c) 2015, Hewlett-Packard Development Company, L.P.<BR>
#  (C) Copyright 2016 Hewlett Packard Enterprise Development LP<BR>
#  Copyright (c) Microsoft Corporation
#
#  SPDX-License-Identifier: BSD-2-Clause-Patent
#
#
# Increase this version tag any time you want user to get warning about updating this
# file in the Conf dir.  By default it does not do update existing conf dirs.
#
# 2.00 - Initial version with changes for CI
#      - Change VS2019 and VS2017 to use PyTool Values for path
#      - Change RC path to use plugin
#      - Change GCC5 for AARCH64 to use ENV instead of DEF
#      - Add path sep for nasm prefix
#


IDENTIFIER = Default TOOL_CHAIN_CONF

# common path macros
DEFINE VS2008_BIN      = ENV(VS2008_PREFIX)
'''


class TestConfMgmt(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        self.test_dir = None
        super().__init__(*args, **kwargs)

    def prep_workspace(self):
        self.clean_workspace()
        self.test_dir = tempfile.mkdtemp()

    def clean_workspace(self):
        if self.test_dir is None:
            return
        if os.path.isdir(self.test_dir):
            shutil.rmtree(self.test_dir)
            self.test_dir = None

    def make_one_file(self, filepath, version):
        with open(filepath, "w") as f:
            f.write(test_file_text)
            f.write(f"#!VERSION={version}")
            logging.debug(f"Made file {filepath} with version {version}")

    def make_edk2_files_dir(self, rootfolderpath, target_ver, tools_def_ver, build_rules_ver, is_template):
        file_names = ["target", "tools_def", "build_rule"]
        versions = [target_ver, tools_def_ver, build_rules_ver]
        if is_template:
            file_names = [x + ".template" for x in file_names]
            rootfolderpath = os.path.join(rootfolderpath, "Conf")
        else:
            file_names = [x + ".txt" for x in file_names]

        os.makedirs(rootfolderpath, exist_ok=True)
        for i in range(0, len(file_names)):
            self.make_one_file(os.path.join(rootfolderpath, file_names[i]), versions[i])

    def get_target_version(self, conf_folder_path):
        p = os.path.join(conf_folder_path, "target.txt")
        return conf_mgmt.ConfMgmt()._get_version(p)

    def get_toolsdef_version(self, conf_folder_path):
        p = os.path.join(conf_folder_path, "tools_def.txt")
        return conf_mgmt.ConfMgmt()._get_version(p)

    def get_buildrules_version(self, conf_folder_path):
        p = os.path.join(conf_folder_path, "build_rule.txt")
        return conf_mgmt.ConfMgmt()._get_version(p)

    def setUp(self):
        self.prep_workspace()

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        self.clean_workspace()

    def test_init(self):
        conf_mgmt.ConfMgmt()

    def test_no_version_tag(self):
        p = os.path.join(self.test_dir, "test.txt")
        with open(p, "w") as f:
            f.write(test_file_text)
        c = conf_mgmt.ConfMgmt()._get_version(p)
        self.assertEqual(c, "0.0")

    def test_invalid_version(self):
        invalid_versions = ["1.2.3.4", "1.2.3", "Hello.1", "1.jk", "",
                            "Wow", "Unknown"]

        p = os.path.join(self.test_dir, "test.txt")
        for a in invalid_versions:
            self.make_one_file(p, a)
            c = conf_mgmt.ConfMgmt()._get_version(p)
            self.assertEqual(c, "0.0")

    def test_valid_version(self):
        valid_versions = ["1.2", "2.02", "0.1", "22.345", "69594.39939494"]

        p = os.path.join(self.test_dir, "test.txt")
        for a in valid_versions:
            self.make_one_file(p, a)
            c = conf_mgmt.ConfMgmt()._get_version(p)
            self.assertEqual(c, a)

    def test_zero_extend_version(self):
        p = os.path.join(self.test_dir, "test.txt")
        self.make_one_file(p, "2")
        c = conf_mgmt.ConfMgmt()._get_version(p)
        self.assertEqual(c, "2.0")

    def test_is_older_version_true(self):
        # first member is conf
        # second member is temp
        # make second member larger
        versions = [("1.2", "1.3"), ("0.1", "0.11"), ("1", "1.000001")]
        conf = os.path.join(self.test_dir, "conf.txt")
        temp = os.path.join(self.test_dir, "temp.txt")
        for a in versions:
            self.make_one_file(conf, a[0])
            self.make_one_file(temp, a[1])
            c = conf_mgmt.ConfMgmt()._is_older_version(conf, temp)
            self.assertTrue(c)

    def test_is_older_version_false(self):
        # first member is conf
        # second member is temp
        # make second member larger
        versions = [("1.2", "1.2"), ("0.1", "0.09"), ("1", ".99999")]
        conf = os.path.join(self.test_dir, "conf.txt")
        temp = os.path.join(self.test_dir, "temp.txt")
        for a in versions:
            self.make_one_file(conf, a[0])
            self.make_one_file(temp, a[1])
            c = conf_mgmt.ConfMgmt()._is_older_version(conf, temp)
            self.assertFalse(c)

    def test_is_older_no_version_tag_conf(self):
        conf = os.path.join(self.test_dir, "conf.txt")
        temp = os.path.join(self.test_dir, "temp.txt")
        # make file with no version tag
        with open(conf, "w") as f:
            f.write(test_file_text)
        self.make_one_file(temp, "1.0")
        c = conf_mgmt.ConfMgmt()._is_older_version(conf, temp)
        self.assertTrue(c)

    def test_is_older_no_version_tag_template(self):
        conf = os.path.join(self.test_dir, "conf.txt")
        temp = os.path.join(self.test_dir, "temp.txt")
        # make file with no version tag
        with open(temp, "w") as f:
            f.write(test_file_text)
        self.make_one_file(conf, "1.0")
        c = conf_mgmt.ConfMgmt()._is_older_version(conf, temp)
        self.assertFalse(c)

    def test_is_older_no_version_no_template(self):
        conf = os.path.join(self.test_dir, "conf.txt")
        temp = os.path.join(self.test_dir, "temp.txt")
        # make file with no version tag
        self.make_one_file(conf, "1.0")
        c = conf_mgmt.ConfMgmt()._is_older_version(conf, temp)
        self.assertFalse(c)

    def test_populate_when_conf_empty(self):
        conf = os.path.join(self.test_dir, "Conf")
        temp = os.path.join(self.test_dir, "templates")

        self.make_edk2_files_dir(temp, "1.0", "2.0", "3.0", True)
        conf_mgmt.ConfMgmt().populate_conf_dir(conf, False, [temp])

        self.assertEqual("1.0", self.get_target_version(conf))
        self.assertEqual("2.0", self.get_toolsdef_version(conf))
        self.assertEqual("3.0", self.get_buildrules_version(conf))

    def test_no_populate_when_conf_fully_populated(self):
        conf = os.path.join(self.test_dir, "Conf")
        temp = os.path.join(self.test_dir, "Templates")

        self.make_edk2_files_dir(temp, "1.0", "2.0", "3.0", True)
        self.make_edk2_files_dir(conf, "1.1", "2.1", "3.1", False)
        conf_mgmt.ConfMgmt().populate_conf_dir(conf, False, [temp])

        self.assertEqual("1.1", self.get_target_version(conf))
        self.assertEqual("2.1", self.get_toolsdef_version(conf))
        self.assertEqual("3.1", self.get_buildrules_version(conf))

    def test_populate_override_fully_populated(self):
        conf = os.path.join(self.test_dir, "Conf")
        temp = os.path.join(self.test_dir, "Templates")

        self.make_edk2_files_dir(temp, "1.0", "2.0", "3.0", True)
        self.make_edk2_files_dir(conf, "1.1", "2.1", "3.1", False)
        conf_mgmt.ConfMgmt().populate_conf_dir(conf, True, [temp])

        self.assertEqual("1.0", self.get_target_version(conf))
        self.assertEqual("2.0", self.get_toolsdef_version(conf))
        self.assertEqual("3.0", self.get_buildrules_version(conf))

    def test_conf_older_warn_fully_populated(self):
        conf = os.path.join(self.test_dir, "Conf")
        temp = os.path.join(self.test_dir, "Templates")

        self.make_edk2_files_dir(temp, "1.2", "2.2", "3.2", True)
        self.make_edk2_files_dir(conf, "1.1", "2.1", "3.1", False)
        c = conf_mgmt.ConfMgmt()
        c._set_delay_time(0)
        c.populate_conf_dir(conf, False, [temp])

        self.assertEqual("1.1", self.get_target_version(conf))
        self.assertEqual("2.1", self.get_toolsdef_version(conf))
        self.assertEqual("3.1", self.get_buildrules_version(conf))

    def test_plat_template_fully_populated(self):
        conf = os.path.join(self.test_dir, "Conf")
        temp = os.path.join(self.test_dir, "Templates")
        plattemp = os.path.join(self.test_dir, "TemplatesPlat")

        self.make_edk2_files_dir(temp, "1.0", "2.0", "3.0", True)
        self.make_edk2_files_dir(plattemp, "1.2", "2.2", "3.2", True)
        conf_mgmt.ConfMgmt().populate_conf_dir(conf, False, [plattemp, temp])

        self.assertEqual("1.2", self.get_target_version(conf))
        self.assertEqual("2.2", self.get_toolsdef_version(conf))
        self.assertEqual("3.2", self.get_buildrules_version(conf))

    def test_no_templates(self):
        conf = os.path.join(self.test_dir, "Conf")
        temp = os.path.join(self.test_dir, "Templates")

        with self.assertRaises(Exception):
            conf_mgmt.ConfMgmt().populate_conf_dir(conf, False, [temp])


if __name__ == '__main__':
    unittest.main()



## @file capsule_helper_test.py
# This unittest module contains test cases for the capsule_helper module.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import os
import uuid
import unittest
import tempfile

from edk2toollib.uefi.uefi_capsule_header import UefiCapsuleHeaderClass
from edk2toollib.uefi.fmp_capsule_header import FmpCapsuleHeaderClass, FmpCapsuleImageHeaderClass
from edk2toolext.capsule import capsule_helper

DUMMY_OPTIONS = {
    'capsule': {
        'fw_version': '0xDEADBEEF',
        'lsv_version': '0xFEEDF00D',
        'esrt_guid': '00112233-4455-6677-8899-aabbccddeeff',
        'fw_name': 'TEST_FW',
        'fw_version_string': '1.2.3',  # deliberately use 3-part version to exercise version normalization.
        'provider_name': 'TESTER',
        'fw_description': 'TEST FW',
        'fw_integrity_file': "IntegrityFile.bin"
    },
    'signer': {
        'option2': 'value2',
        'option_not': 'orig_value'
    }
}
DUMMY_OPTIONS_FILE_NAME = 'dummy_options_file'
DUMMY_PAYLOAD_FILE_NAME = 'dummy_payload'


class CapsuleSignerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # We'll use the one-time setup to create
        # any temporary test files we'll need.
        cls.temp_dir = tempfile.mkdtemp()
        cls.dummy_payload = os.path.join(cls.temp_dir, DUMMY_PAYLOAD_FILE_NAME + ".bin")

        with open(cls.dummy_payload, 'wb') as dummy_file:
            dummy_file.write(b'DEADBEEF')

    def test_should_pass_wrapped_blob_to_signing_module(self):
        dummy_payload = b'This_Is_My_Sample_Payload,ThereAreManyLikeIt;This One Is Mine'

        class DummySigner(object):
            @classmethod
            def sign(cls, data, signature_options, signer_options):
                self.assertTrue(dummy_payload in data)

        capsule_helper.build_capsule(dummy_payload, DUMMY_OPTIONS['capsule'], DummySigner, DUMMY_OPTIONS['signer'])

    def test_should_pass_signer_options_to_signing_module(self):
        class DummySigner(object):
            @classmethod
            def sign(cls, data, signature_options, signer_options):
                self.assertEqual(signer_options, DUMMY_OPTIONS['signer'])

        capsule_helper.build_capsule(b'030303', DUMMY_OPTIONS['capsule'], DummySigner, DUMMY_OPTIONS['signer'])

    # def test_should_be_able_to_generate_a_production_equivalent_capsule(self):
    #     with open(BUILD_CAPSULE_BINARY_PATH, 'rb') as data_file:
    #         capsule_data = data_file.read()

    #     capsule_options = {
    #         "esrt_guid": "80ddc468-57a0-43e5-9594-8ba2ce5c342e",
    #         "fw_version": "0x7fff000",
    #         "lsv_version": "0x1"
    #     }
    #     signer_options = {
    #         'key_file': TEST_CAPSULE_SIGNER_1,
    #         'eku': "1.3.6.1.4.1.311.76.9.1.36"
    #     }
    #     wdk_signer = signing_helper.get_signer(signing_helper.SIGNTOOL_SIGNER)

    #     final_capsule = capsule_helper.build_capsule(capsule_data, capsule_options, wdk_signer, signer_options)

    #     with open(TEST_CAPSULE_PATH_1, 'rb') as comparison_file:
    #         comparison_data = comparison_file.read()

    #     self.assertEqual(final_capsule.Encode(), comparison_data)

    # def test_should_be_able_to_update_the_guid_in_place(self):
    #     with open(BUILD_CAPSULE_BINARY_PATH, 'rb') as data_file:
    #         capsule_data = data_file.read()

    #     capsule_options = {
    #         "esrt_guid": "3624cd98-bdb6-461b-84a3-4f4853efc7e3",
    #         "fw_version": "0x7fff000",
    #         "lsv_version": "0x1"
    #     }
    #     signer_options = {
    #         'key_file': TEST_CAPSULE_SIGNER_1,
    #         'eku': "1.3.6.1.4.1.311.76.9.1.36"
    #     }
    #     wdk_signer = signing_helper.get_signer(signing_helper.SIGNTOOL_SIGNER)

    #     final_capsule = capsule_helper.build_capsule(capsule_data, capsule_options, wdk_signer, signer_options)

    #     with open(os.path.join(TEMP_CAPSULE_DIRECTORY_PATH, 'Capsule1.bin'), 'wb') as out_file:
    #         out_file.write(final_capsule.Encode())
    #         final_capsule.DumpInfo()

    #     fmp_capsule_image_header = final_capsule.FmpCapsuleHeader.GetFmpCapsuleImageHeader(0)
    #     fmp_capsule_image_header.UpdateImageTypeId = uuid.UUID('80ddc468-57a0-43e5-9594-8ba2ce5c342e')
    #     # WORKAROUND for library bug.
    #     final_capsule.FmpCapsuleHeader._ItemOffsetList = []

    #     with open(os.path.join(TEMP_CAPSULE_DIRECTORY_PATH, 'Capsule2.bin'), 'wb') as out_file:
    #         out_file.write(final_capsule.Encode())
    #         final_capsule.DumpInfo()

    #     self.assertFalse(True)


class FileGenerationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # We'll use the one-time setup to create
        # any temporary test files we'll need.
        cls.temp_dir = tempfile.mkdtemp()
        cls.dummy_payload = os.path.join(cls.temp_dir, DUMMY_PAYLOAD_FILE_NAME + ".bin")

        with open(cls.dummy_payload, 'wb') as dummy_file:
            dummy_file.write(b'DEADBEEF')

    def test_should_be_able_to_save_a_capsule(self):
        fmp_capsule_image_header = FmpCapsuleImageHeaderClass()
        fmp_capsule_image_header.UpdateImageTypeId = uuid.UUID(DUMMY_OPTIONS['capsule']['esrt_guid'])
        fmp_capsule_image_header.UpdateImageIndex = 1

        fmp_capsule_header = FmpCapsuleHeaderClass()
        fmp_capsule_header.AddFmpCapsuleImageHeader(fmp_capsule_image_header)

        uefi_capsule_header = UefiCapsuleHeaderClass()
        uefi_capsule_header.FmpCapsuleHeader = fmp_capsule_header
        uefi_capsule_header.PersistAcrossReset = True
        uefi_capsule_header.InitiateReset = True

        capsule_file_path = capsule_helper.save_capsule(uefi_capsule_header, DUMMY_OPTIONS['capsule'], self.temp_dir)

        # Now read the data and check for the GUID.
        with open(capsule_file_path, 'rb') as capsule_file:
            capsule_bytes = capsule_file.read()

        self.assertTrue(uuid.UUID(DUMMY_OPTIONS['capsule']['esrt_guid']).bytes_le in capsule_bytes)

    def test_should_be_able_to_generate_windows_files(self):
        inf_file_path = capsule_helper.create_inf_file(DUMMY_OPTIONS['capsule'], self.temp_dir)
        self.assertTrue(os.path.isfile(inf_file_path))

    @unittest.skip("test fails in unittest environment. need to debug")
    def test_should_be_able_to_generate_cat(self):
        cat_file_path = capsule_helper.create_cat_file(DUMMY_OPTIONS['capsule'], self.temp_dir)
        self.assertTrue(os.path.isfile(cat_file_path))


class MultiNodeFileGenerationTest(unittest.TestCase):

    @staticmethod
    def buildPayload(esrt):
        fmp_capsule_image_header = FmpCapsuleImageHeaderClass()
        fmp_capsule_image_header.UpdateImageTypeId = uuid.UUID(esrt)
        fmp_capsule_image_header.UpdateImageIndex = 1

        fmp_capsule_header = FmpCapsuleHeaderClass()
        fmp_capsule_header.AddFmpCapsuleImageHeader(fmp_capsule_image_header)

        uefi_capsule_header = UefiCapsuleHeaderClass()
        uefi_capsule_header.FmpCapsuleHeader = fmp_capsule_header
        uefi_capsule_header.PersistAcrossReset = True
        uefi_capsule_header.InitiateReset = True

        return uefi_capsule_header

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.mkdtemp()
        cls.temp_output_dir = tempfile.mkdtemp()

        cls.capsule = capsule_helper.Capsule(
            version_string="1.2.3",
            name="TEST_FW",
            provider_name="Tester",
        )

        cls.capsule.payloads.append(
            capsule_helper.CapsulePayload(
                cls.buildPayload("ea5c13fe-cac9-4fd7-ac30-37709bd668f2"),
                "test1.bin",
                uuid.UUID("ea5c13fe-cac9-4fd7-ac30-37709bd668f2"),
                0xDEADBEEF,
                "TEST FW"
            )
        )

        cls.capsule.payloads.append(
            capsule_helper.CapsulePayload(
                cls.buildPayload("43e67b4e-b2f1-4891-9ff2-a6acd9c74cbd"),
                "test2.bin",
                uuid.UUID("43e67b4e-b2f1-4891-9ff2-a6acd9c74cbd"),
                0xDEADBEEF,
                "TEST FW"
            )
        )

    def test_should_be_able_to_save_a_multi_node_capsule(self):

        capsule_file_path = capsule_helper.save_multinode_capsule(self.capsule, self.temp_output_dir)

        # make sure all the files we expect got created
        for payload in self.capsule.payloads:
            payload_file = os.path.join(capsule_file_path, payload.payload_filename)
            self.assertTrue(os.path.isfile(payload_file))
            with open(payload_file, 'rb') as fh:
                capsule_bytes = fh.read()
            self.assertIn(payload.esrt_guid.bytes_le, capsule_bytes)

    def test_should_be_able_to_save_a_multi_node_capsule_with_integrity(self):

        self.capsule.payloads[0].integrity_data = uuid.UUID("ea5c13fe-cac9-4fd7-ac30-37709bd668f2").bytes
        self.capsule.payloads[0].integrity_filename = "integrity1.bin"

        self.capsule.payloads[1].integrity_data = uuid.UUID("43e67b4e-b2f1-4891-9ff2-a6acd9c74cbd").bytes
        self.capsule.payloads[1].integrity_filename = "integrity2.bin"

        capsule_file_path = capsule_helper.save_multinode_capsule(self.capsule, self.temp_output_dir)

        for payload in self.capsule.payloads:
            payload_file = os.path.join(capsule_file_path, payload.payload_filename)
            self.assertTrue(os.path.isfile(payload_file))
            with open(payload_file, 'rb') as fh:
                capsule_bytes = fh.read()
            self.assertIn(payload.esrt_guid.bytes_le, capsule_bytes)

            integrityFile = os.path.join(capsule_file_path, payload.integrity_filename)
            self.assertTrue(os.path.isfile(integrityFile))
            with open(integrityFile, 'rb') as fh:
                integrity_bytes = fh.read()
            self.assertIn(payload.integrity_data, integrity_bytes)

        self.capsule.payloads[0].integrity_data = None
        self.capsule.payloads[0].integrity_filename = None

        self.capsule.payloads[1].integrity_data = None
        self.capsule.payloads[1].integrity_filename = None

    def test_should_be_able_to_generate_multi_node_inf_file(self):

        inf_file_path = capsule_helper.create_multinode_inf_file(self.capsule, self.temp_output_dir)
        self.assertTrue(os.path.isfile(inf_file_path))


if __name__ == '__main__':
    unittest.main()



# @file signing_helper.py
# This module contains code to help with the selection and loading of a signer module.
# These signer modules can be built-in to the edk2toolext module, loaded from other Pip
# or Python modules that are on the current pypath, or passed as a file path to a
# local Python module that should be dynamically loaded.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help with the selection and loading of a signer module.

These signer modules can be built-in to the edk2toolext module, loaded from
other Pip or Python modules that are on the current pypath, or passed as a
file path to a local Python module that should be dynamically loaded.
"""

import importlib

from edk2toolext.capsule import signtool_signer
from edk2toollib.utility_functions import import_module_by_file_name

# Valid types.
PYOPENSSL_SIGNER = 'pyopenssl'
SIGNTOOL_SIGNER = 'signtool'
PYPATH_MODULE_SIGNER = 'pymodule'
LOCAL_MODULE_SIGNER = 'local_module'


def get_signer(type, specifier=None):
    """Load a signer module based on the arguments.

    if type is PYPATH_MODULE_SIGNER, the specifier should be the python module.
    i.e. edk2toolext.capsule.pyopenssl_signer

    if the type is LOCAL_MODULE_SIGNER, the specifier should be a filesystem
    path to a Python module that can be loaded as the signer.

    Args:
        type (str): PYOPENSSL_SIGNER, SIGNTOOL_SIGNER, PYPATH_MODULE_SIGNER, LOCAL_MODULE_SIGNER
        specifier (module): python module to import
    """
    if type == PYOPENSSL_SIGNER:
        try:
            from edk2toolext.capsule import pyopenssl_signer
            return pyopenssl_signer
        except ModuleNotFoundError:
            raise RuntimeError('PyOpenSsl Signer failed to load. Do you have pyopenssl installed?')
    elif type == SIGNTOOL_SIGNER:
        return signtool_signer
    elif type == PYPATH_MODULE_SIGNER:
        return importlib.import_module(specifier)
    elif type == LOCAL_MODULE_SIGNER:
        return import_module_by_file_name(specifier)
    else:
        return None



# @file edk2_platform_build
# Invocable class that does a build.
# Needs a child of UefiBuilder for pre/post build steps.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Invocable class that does a build.

Contains a BuildSettingsManager that must be subclassed in a build settings
file, along with a UefiBuilder subclass. This provides platform specific
information to the Edk2PlatformBuild invocable while allowing the invocable
itself to remain platform agnostic.
"""
import os
import sys
import logging
from edk2toolext import edk2_logging
from edk2toolext.environment import plugin_manager
from edk2toolext.environment.plugintypes.uefi_helper_plugin import HelperFunctions
from edk2toolext.environment import self_describing_environment
from edk2toolext.environment.uefi_build import UefiBuilder
from edk2toolext.edk2_invocable import Edk2Invocable, Edk2InvocableSettingsInterface
from edk2toollib.utility_functions import locate_class_in_module
from edk2toollib.uefi.edk2.path_utilities import Edk2Path


class BuildSettingsManager(Edk2InvocableSettingsInterface):
    """Platform specific settings for Edk2PlatformBuild.

    Provides information necessary for `stuart_build.exe`
    or `edk2_platform_build.py` to successfully execute.

    Example: Example: Overriding BuildSettingsManager
        ```python
        from edk2toolext.invocables.edk2_platform_build import BuildSettingsManager
        class PlatformManager(BuildSettingsManager):
            def GetName(self) -> str:
                return "QemuQ35"
        ```
    """

    def GetName(self) -> str:
        """Get the name of the repo, platform, or product being build.

        TIP: Optional Override in subclass

        Returns:
            (str): Name of the repo, platform, or product
        """
        return None


class Edk2PlatformBuild(Edk2Invocable):
    """Invocable that performs some environment setup,Imports UefiBuilder and calls go."""

    def AddCommandLineOptions(self, parserObj):
        """Adds command line options to the argparser."""
        # PlatformSettings could also be a subclass of UefiBuilder, who knows!
        if isinstance(self.PlatformSettings, UefiBuilder):
            self.PlatformBuilder = self.PlatformSettings
        else:
            try:
                # if it's not, we will try to find it in the module that was originally provided.
                self.PlatformBuilder = locate_class_in_module(self.PlatformModule, UefiBuilder)()
            except (TypeError):
                raise RuntimeError(f"UefiBuild not found in module:\n{dir(self.PlatformModule)}")

        self.PlatformBuilder.AddPlatformCommandLineOptions(parserObj)

    def RetrieveCommandLineOptions(self, args):
        """Retrieve command line options from the argparser."""
        self.PlatformBuilder.RetrievePlatformCommandLineOptions(args)

    def GetSettingsClass(self):
        """Returns the BuildSettingsManager class.

        WARNING: CiSetupSettingsManager must be subclassed in your platform settings file.
        """
        return BuildSettingsManager

    def GetLoggingFileName(self, loggerType):
        """Returns the filename of where the logs for the Edk2PlatformBuild invocable are stored in."""
        name = self.PlatformSettings.GetName()
        if name is not None:
            return f"BUILDLOG_{name}"
        return "BUILDLOG"

    def Go(self):
        """Executes the core functionality of the Edk2PlatformBuild invocable."""
        logging.info("Running Python version: " + str(sys.version_info))

        Edk2PlatformBuild.collect_python_pip_info()

        (build_env, shell_env) = self_describing_environment.BootstrapEnvironment(
            self.GetWorkspaceRoot(), self.GetActiveScopes(), self.GetSkippedDirectories())

        # Bind our current execution environment into the shell vars.
        ph = os.path.dirname(sys.executable)
        if " " in ph:
            ph = '"' + ph + '"'
        shell_env.set_shell_var("PYTHON_HOME", ph)
        # PYTHON_COMMAND is required to be set for using edk2 python builds.
        # todo: work with edk2 to remove the bat file and move to native python calls
        pc = sys.executable
        if " " in pc:
            pc = '"' + pc + '"'
        shell_env.set_shell_var("PYTHON_COMMAND", pc)

        # Load plugins
        logging.log(edk2_logging.SECTION, "Loading Plugins")
        pm = plugin_manager.PluginManager()
        failedPlugins = pm.SetListOfEnvironmentDescriptors(
            build_env.plugins)
        if failedPlugins:
            logging.critical("One or more plugins failed to load. Halting build.")
            for a in failedPlugins:
                logging.error("Failed Plugin: {0}".format(a["name"]))
            raise Exception("One or more plugins failed to load.")

        helper = HelperFunctions()
        if (helper.LoadFromPluginManager(pm) > 0):
            raise Exception("One or more helper plugins failed to load.")

        # Make a pathobj so we can normalize and validate the workspace
        # and packages path.  The Settings Manager can return absolute or
        # relative paths
        pathobj = Edk2Path(self.GetWorkspaceRoot(), self.GetPackagesPath())
        #
        # Now we can actually kick off a build.
        #
        logging.log(edk2_logging.SECTION, "Kicking off build")
        ret = self.PlatformBuilder.Go(pathobj.WorkspacePath,
                                      os.pathsep.join(pathobj.PackagePathList),
                                      helper, pm)
        logging.log(edk2_logging.SECTION, f"Log file is located at: {self.log_filename}")
        return ret


def main():
    """Entry point invoke Edk2PlatformBuild."""
    Edk2PlatformBuild().Invoke()



# @file test_nuget.py
# This contains unit tests for the nuget binary wrapper
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import tempfile
import unittest
from edk2toolext.bin import nuget


class Test_nuget(unittest.TestCase):

    def test_can_download_nuget(self):
        test_dir = tempfile.mkdtemp()
        nuget_path = os.path.join(test_dir, "NuGet.exe")
        self.assertFalse(os.path.exists(nuget_path))
        nuget.DownloadNuget(test_dir)
        self.assertTrue(os.path.exists(nuget_path))

    def test_checks_nuget_sha(self):
        test_dir = tempfile.mkdtemp()
        nuget_path = os.path.join(test_dir, "NuGet.exe")
        file = open(nuget_path, "w")
        file.write("This is not the nuget that you are looking for\n. :)")
        file.close()
        with self.assertRaises(RuntimeError):
            nuget.DownloadNuget(test_dir)



# @file test_shell_environment.py
# Unit test suite for the ShellEnvironment class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import os
import sys
import unittest
import edk2toolext.environment.shell_environment as SE


class TestShellEnvironmentAssumptions(unittest.TestCase):

    def test_shell_should_be_a_singleton(self):
        shell_a = SE.ShellEnvironment()
        shell_b = SE.ShellEnvironment()
        self.assertIs(shell_a, shell_b, "two instances of ShellEnvironment should be identical")

    def test_shell_tests_need_to_be_able_to_clear_singleton(self):
        # This is not currently achievable, and may never be achievable.
        pass

    def test_shell_should_always_have_an_initial_checkpoint(self):
        shell_env = SE.ShellEnvironment()
        self.assertTrue((len(shell_env.checkpoints) > 0),
                        "a new instance of ShellEnvironment should have at least one checkpoint")


class TestBasicEnvironmentManipulation(unittest.TestCase):

    def test_can_set_os_vars(self):
        shell_env = SE.ShellEnvironment()
        # Remove the test var, if it exists.
        os.environ.pop("SE-TEST-VAR-1", None)
        # Set a new value and get it directly from the environment.
        new_value = 'Dummy'
        shell_env.set_shell_var('SE-TEST-VAR-1', new_value)
        self.assertEqual(os.environ['SE-TEST-VAR-1'], new_value)

    def test_can_get_os_vars(self):
        shell_env = SE.ShellEnvironment()
        new_value = 'Dummy2'
        shell_env.set_shell_var('SE-TEST-VAR-2', new_value)
        self.assertEqual(shell_env.get_shell_var('SE-TEST-VAR-2'), new_value)

    def test_set_path_string(self):
        shell_env = SE.ShellEnvironment()

        # Test pass 1.
        testpath_elems = ['MY_PATH']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_path(testpath_string)
        self.assertEqual(os.environ['PATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_path, "the active path should contain all elements")

        # Test pass 2.
        testpath_elems = ['/bin/bash', 'new_path', '/root']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_path(testpath_string)
        self.assertEqual(os.environ['PATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_path, "the active path should contain all elements")

    def test_set_path_elements(self):
        shell_env = SE.ShellEnvironment()

        # Test pass 1.
        testpath_elems = ['MY_PATH']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_path(testpath_elems)
        self.assertEqual(os.environ['PATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_path, "the active path should contain all elements")

        # Test pass 2.
        testpath_elems = ['/bin/bash', 'new_path', '/root']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_path(testpath_elems)
        self.assertEqual(os.environ['PATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_path, "the active path should contain all elements")

    def test_set_pypath_string(self):
        shell_env = SE.ShellEnvironment()

        # Test pass 1.
        testpath_elems = ['MY_PATH']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_pypath(testpath_string)
        self.assertEqual(os.environ['PYTHONPATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_pypath, "the active path should contain all elements")
            self.assertIn(elem, sys.path, "the sys path should contain all elements")

        # Test pass 2.
        testpath_elems = ['/bin/bash', 'new_path', '/root']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_pypath(testpath_string)
        self.assertEqual(os.environ['PYTHONPATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_pypath, "the active path should contain all elements")
            self.assertIn(elem, sys.path, "the sys path should contain all elements")

    def test_set_pypath_elements(self):
        shell_env = SE.ShellEnvironment()

        # Test pass 1.
        testpath_elems = ['MY_PATH']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_pypath(testpath_elems)
        self.assertEqual(os.environ['PYTHONPATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_pypath, "the active path should contain all elements")
            self.assertIn(elem, sys.path, "the sys path should contain all elements")

        # Test pass 2.
        testpath_elems = ['/bin/bash', 'new_path', '/root']
        testpath_string = os.pathsep.join(testpath_elems)
        shell_env.set_pypath(testpath_elems)
        self.assertEqual(os.environ['PYTHONPATH'], testpath_string, "the final string should be correct")
        for elem in testpath_elems:
            self.assertIn(elem, shell_env.active_pypath, "the active path should contain all elements")
            self.assertIn(elem, sys.path, "the sys path should contain all elements")

    def test_insert_append_remove_replace_path(self):
        shell_env = SE.ShellEnvironment()

        # Start with a known PATH
        mid_elem = 'MIDDLEPATH'
        shell_env.set_path(mid_elem)
        self.assertEqual(1, len(shell_env.active_path))
        self.assertIn(mid_elem, shell_env.active_path)
        # Add an element to the end.
        end_elem = 'ENDPATH'
        shell_env.append_path(end_elem)
        # Add an element to the beginning.
        start_elem = 'STARTPATH'
        shell_env.insert_path(start_elem)

        # Test for the realities.
        self.assertEqual(3, len(shell_env.active_path))
        self.assertEqual(shell_env.active_path[0], start_elem)
        self.assertEqual(shell_env.active_path[1], mid_elem)
        self.assertEqual(shell_env.active_path[2], end_elem)
        for elem in (start_elem, mid_elem, end_elem):
            self.assertIn(elem, os.environ["PATH"])

        # Test inserting existing elements as a method to move them front
        shell_env.insert_path(end_elem)
        self.assertEqual(3, len(shell_env.active_path))
        self.assertEqual(shell_env.active_path[0], end_elem)
        self.assertEqual(shell_env.active_path[1], start_elem)
        self.assertEqual(shell_env.active_path[2], mid_elem)

        # Test appending existing elements as a method to move to end
        shell_env.append_path(end_elem)
        self.assertEqual(3, len(shell_env.active_path))
        self.assertEqual(shell_env.active_path[0], start_elem)
        self.assertEqual(shell_env.active_path[1], mid_elem)
        self.assertEqual(shell_env.active_path[2], end_elem)

        # Test replacing an element on the path
        new_mid_elem = 'NEWMIDDLEPATH'
        shell_env.replace_path_element(mid_elem, new_mid_elem)
        self.assertEqual(shell_env.active_path[1], new_mid_elem)

        # Test replacing an element that doesn't exist
        old_path = shell_env.active_path
        shell_env.replace_path_element("PATH1", "PATH2")
        self.assertEqual(old_path, shell_env.active_path)

        # Test that removing an element works as expected
        shell_env.remove_path_element(new_mid_elem)
        self.assertNotIn(new_mid_elem, shell_env.active_path)

        # Test that calling Set Var can replace the path
        shell_env.set_shell_var("PATH", "a" + os.pathsep + "b")
        self.assertEqual(len(os.environ["PATH"].split(os.pathsep)), 2)

        # Test that calling Set Var can replace the path and is case insensitive
        shell_env.set_shell_var("pAtH", "c" + os.pathsep + "d" + os.pathsep + "e")
        self.assertEqual(len(os.environ["PATH"].split(os.pathsep)), 3)

    def test_insert_append_remove_replace_pypath(self):
        shell_env = SE.ShellEnvironment()

        # Start with a known PATH
        mid_elem = 'MIDDLEPATH'
        shell_env.set_pypath(mid_elem)
        self.assertEqual(1, len(shell_env.active_pypath))
        self.assertIn(mid_elem, shell_env.active_pypath)
        # Add an element to the end.
        end_elem = 'ENDPATH'
        shell_env.append_pypath(end_elem)
        # Add an element to the beginning.
        start_elem = 'STARTPATH'
        shell_env.insert_pypath(start_elem)

        # Test for the realities.
        self.assertEqual(3, len(shell_env.active_pypath))
        self.assertEqual(shell_env.active_pypath[0], start_elem)
        self.assertEqual(shell_env.active_pypath[1], mid_elem)
        self.assertEqual(shell_env.active_pypath[2], end_elem)
        for elem in (start_elem, mid_elem, end_elem):
            self.assertIn(elem, os.environ["PYTHONPATH"])
            self.assertIn(elem, sys.path)

        # Test replacing an element on the pypath
        new_mid_elem = 'NEWMIDDLEPATH'
        shell_env.replace_pypath_element(mid_elem, new_mid_elem)
        self.assertEqual(shell_env.active_pypath[1], new_mid_elem)

        # Test replacing an element that doesn't exist
        old_pypath = shell_env.active_pypath
        shell_env.replace_pypath_element("PATH1", "PATH2")
        self.assertEqual(old_pypath, shell_env.active_pypath)

        # Test that removing an element works as expected
        shell_env.remove_pypath_element(new_mid_elem)
        self.assertNotIn(new_mid_elem, shell_env.active_pypath)

    def test_can_set_and_get_build_vars(self):
        shell_env = SE.ShellEnvironment()

        var_name = 'SE-TEST-VAR-3'
        var_data = 'Dummy3'
        # Make sure it doesn't exist beforehand.
        self.assertIs(shell_env.get_build_var(var_name), None, "test var should not exist before creation")
        shell_env.set_build_var(var_name, var_data)
        self.assertEqual(shell_env.get_build_var(var_name), var_data, "get var data should match set var data")

    def test_set_build_vars_should_default_overrideable(self):
        shell_env = SE.ShellEnvironment()

        var_name = 'SE_TEST_VAR_4'
        var_data = 'NewData1'
        var_data2 = 'NewerData1'

        self.assertIs(shell_env.get_build_var(var_name), None, "test var should not exist before creation")
        shell_env.set_build_var(var_name, var_data)
        shell_env.set_build_var(var_name, var_data2)

        self.assertEqual(shell_env.get_build_var(var_name), var_data2)


class TestShellEnvironmenCheckpoints(unittest.TestCase):

    def setUp(self):
        # Grab the singleton and restore the initial checkpoint.
        shell_env = SE.ShellEnvironment()
        shell_env.restore_initial_checkpoint()
        # For testing, purge all checkpoints each time.
        shell_env.checkpoints = [shell_env.checkpoints[SE.ShellEnvironment.INITIAL_CHECKPOINT]]

    def test_restore_initial_checkpoint_should_erase_changes(self):
        shell_env = SE.ShellEnvironment()

        # Check to make sure the change doesn't exist.
        test_path_change = '/SE/TEST/PATH/1'
        self.assertNotIn(test_path_change, shell_env.active_path, "starting condition should not have the test change")

        # Make the change and verify.
        shell_env.append_path(test_path_change)
        self.assertIn(test_path_change, shell_env.active_path)

        # Add a shell_var while we're at it.
        self.assertEqual(shell_env.get_shell_var('i_should_not_exist'), None)
        shell_env.set_shell_var('i_should_not_exist', 'a_value')
        self.assertEqual(shell_env.get_shell_var('i_should_not_exist'), 'a_value')

        # Restore initial checkpoint and verify change is gone.
        shell_env.restore_initial_checkpoint()
        self.assertNotIn(test_path_change, shell_env.active_path, "restoring checkpoint should remove test change")
        self.assertEqual(shell_env.get_shell_var('i_should_not_exist'), None)

    def test_checkpoint_indices_should_be_unique(self):
        shell_env = SE.ShellEnvironment()
        shell_env.append_path('/SE/TEST/PATH/1')
        check_point1 = shell_env.checkpoint()
        shell_env.append_path('/SE/TEST/PATH/2')
        check_point2 = shell_env.checkpoint()

        self.assertNotEqual(check_point1, SE.ShellEnvironment.INITIAL_CHECKPOINT)
        self.assertNotEqual(check_point2, SE.ShellEnvironment.INITIAL_CHECKPOINT)
        self.assertNotEqual(check_point1, check_point2)

    def test_restore_new_checkpoint_should_contain_new_changes(self):
        shell_env = SE.ShellEnvironment()

        # Check to make sure the change doesn't exist.
        test_path_change = '/SE/TEST/PATH/3'
        self.assertNotIn(test_path_change, shell_env.active_path, "starting condition should not have the test change")

        # Make the change and checkpoint.
        shell_env.append_path(test_path_change)
        self.assertIn(test_path_change, shell_env.active_path)
        check_point1 = shell_env.checkpoint()

        # Restore initial checkpoint and verify change is gone.
        shell_env.restore_initial_checkpoint()
        self.assertNotIn(test_path_change, shell_env.active_path,
                         "restoring initial checkpoint should remove test change")

        # Restore new checkpoint and verify change is back.
        shell_env.restore_checkpoint(check_point1)
        self.assertIn(test_path_change, shell_env.active_path, "restoring new checkpoint should restore test change")

    def test_checkpointed_objects_should_behave_correctly(self):
        shell_env = SE.ShellEnvironment()

        # This test is to make sure that pass-by-reference elements don't persist unexpectedly.

        test_var1_name = 'SE_TEST_VAR_3'
        test_var1_data = 'MyData1'
        test_var1_data2 = 'RevisedData1'
        test_var1_data3 = 'MoreRevisedData1'

        test_var2_name = 'SE_TEST_VAR_4'
        test_var2_data = 'MyData2'

        # Set the first data and make a checkpoint.
        shell_env.set_build_var(test_var1_name, test_var1_data)
        check_point1 = shell_env.checkpoint()

        # Update previous value and set second data. Then checkpoint.
        shell_env.set_build_var(test_var1_name, test_var1_data2)
        shell_env.set_build_var(test_var2_name, test_var2_data)
        check_point2 = shell_env.checkpoint()

        # Restore the first checkpoint and verify values.
        shell_env.restore_checkpoint(check_point1)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data)
        self.assertIs(shell_env.get_build_var(test_var2_name), None)

        # Make a change to be tested later.
        shell_env.set_build_var(test_var1_name, test_var1_data3)

        # Restore the second checkpoint and verify values.
        shell_env.restore_checkpoint(check_point2)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data2)
        self.assertEqual(shell_env.get_build_var(test_var2_name), test_var2_data)

        # Restore the first checkpoint again and make sure original value still stands.
        shell_env.restore_checkpoint(check_point1)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data)


class TestShellEnvironmenSpecialBuildVars(unittest.TestCase):

    def setUp(self):
        # Grab the singleton and restore the initial checkpoint.
        shell_env = SE.ShellEnvironment()
        shell_env.restore_initial_checkpoint()
        # For testing, purge all checkpoints each time.
        shell_env.checkpoints = [shell_env.checkpoints[SE.ShellEnvironment.INITIAL_CHECKPOINT]]

    def test_get_build_vars_should_update_vars(self):
        shell_env = SE.ShellEnvironment()
        build_vars = SE.GetBuildVars()

        test_var_name = 'SE_TEST_VAR_4'
        test_var_data = 'NewData1'

        build_vars.SetValue(test_var_name, test_var_data, 'random set')

        self.assertEqual(shell_env.get_build_var(test_var_name), test_var_data)

    def test_special_build_vars_should_default_non_overrideable(self):
        shell_env = SE.ShellEnvironment()
        build_vars = SE.GetBuildVars()

        test_var_name = 'SE_TEST_VAR_4'
        test_var_data = 'NewData1'
        test_var_data2 = 'NewerData1'

        build_vars.SetValue(test_var_name, test_var_data, 'random set')
        build_vars.SetValue(test_var_name, test_var_data2, 'another random set')

        self.assertEqual(shell_env.get_build_var(test_var_name), test_var_data)

    def test_special_build_vars_should_always_update_current(self):
        shell_env = SE.ShellEnvironment()
        build_vars = SE.GetBuildVars()

        test_var1_name = 'SE_TEST_VAR_update_current1'
        test_var1_data = 'NewData1'
        test_var1_data2 = 'NewerData1'

        test_var2_name = 'SE_TEST_VAR_update_current2'
        test_var2_data = 'NewData2'

        # Make a change and checkpoint.
        build_vars.SetValue(test_var1_name, test_var1_data, 'var1 set', overridable=True)
        shell_env.checkpoint()

        # Make a couple more changes.
        build_vars.SetValue(test_var1_name, test_var1_data2, 'var1 set', overridable=True)
        build_vars.SetValue(test_var2_name, test_var2_data, 'var2 set', overridable=True)

        # Make sure that the newer changes are valid.
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data2)
        self.assertEqual(shell_env.get_build_var(test_var2_name), test_var2_data)

    def test_special_build_vars_should_be_checkpointable(self):
        shell_env = SE.ShellEnvironment()
        build_vars = SE.GetBuildVars()

        # This test is basically a rehash of the object checkpointing test,
        # but this time with the special vars.

        test_var1_name = 'SE_TEST_VAR_3'
        test_var1_data = 'MyData1'
        test_var1_data2 = 'RevisedData1'
        test_var1_data3 = 'MoreRevisedData1'

        test_var2_name = 'SE_TEST_VAR_4'
        test_var2_data = 'MyData2'

        # Set the first data and make a checkpoint.
        build_vars.SetValue(test_var1_name, test_var1_data, 'var1 set', overridable=True)
        check_point1 = shell_env.checkpoint()

        # Update previous value and set second data. Then checkpoint.
        build_vars.SetValue(test_var1_name, test_var1_data2, 'var1 set', overridable=True)
        build_vars.SetValue(test_var2_name, test_var2_data, 'var2 set', overridable=True)
        check_point2 = shell_env.checkpoint()

        # Restore the first checkpoint and verify values.
        shell_env.restore_checkpoint(check_point1)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data)
        self.assertIs(shell_env.get_build_var(test_var2_name), None)

        # Make a change to be tested later.
        build_vars.SetValue(test_var1_name, test_var1_data3, 'var1 set', overridable=True)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data3,
                         'even after restore, special build vars should always update current')

        # Restore the second checkpoint and verify values.
        shell_env.restore_checkpoint(check_point2)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data2)
        self.assertEqual(shell_env.get_build_var(test_var2_name), test_var2_data)

        # Restore the first checkpoint again and make sure original value still stands.
        shell_env.restore_checkpoint(check_point1)
        self.assertEqual(shell_env.get_build_var(test_var1_name), test_var1_data)


if __name__ == '__main__':
    unittest.main()



# @file test_uefi_build.py
# Unit test suite for the UefiB class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext.environment import uefi_build
from edk2toolext.environment.plugintypes import uefi_helper_plugin
from edk2toolext.environment.plugin_manager import PluginManager
from edk2toollib.utility_functions import GetHostInfo
import argparse
import tempfile
import os
import stat
from inspect import cleandoc
from edk2toolext.environment import shell_environment


class TestUefiBuild(unittest.TestCase):

    def setUp(self):
        self.WORKSPACE = tempfile.mkdtemp()
        TestUefiBuild.create_min_uefi_build_tree(self.WORKSPACE)
        pass

    def tearDown(self):
        shell_environment.GetEnvironment().restore_initial_checkpoint()
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_init(self):
        builder = uefi_build.UefiBuilder()
        self.assertIsNotNone(builder)

    @classmethod
    def write_to_file(cls, path, contents, close=True):
        f = open(path, "w")
        f.writelines(contents)
        if close:
            f.close()

    @classmethod
    def create_min_uefi_build_tree(cls, root):
        conf_folder = os.path.join(root, "Conf")
        os.makedirs(conf_folder)
        target_path = os.path.join(conf_folder, "target.template")
        TestUefiBuild.write_to_file(target_path, ["ACTIVE_PLATFORM = Test.dsc\n",
                                                  "TOOL_CHAIN_TAG = test\n",
                                                  "TARGET = DEBUG\n"])
        tools_path = os.path.join(conf_folder, "tools_def.template")
        TestUefiBuild.write_to_file(tools_path, ["hello"])
        build_path = os.path.join(conf_folder, "build_rule.template")
        TestUefiBuild.write_to_file(build_path, ["hello"])
        platform_path = os.path.join(root, "Test.dsc")
        TestUefiBuild.write_to_file(platform_path, ["[Defines]\n",
                                                    "OUTPUT_DIRECTORY = Build"])

    def test_commandline_options(self):
        builder = uefi_build.UefiBuilder()
        parserObj = argparse.ArgumentParser()
        builder.AddPlatformCommandLineOptions(parserObj)
        args = [
            ["--CLEAN", "--SKIPBUILD"],
            ["--FLASHONLY"],
            ["--CLEANONLY"],
            ["--FLASHROM"],
            ["--UPDATECONF"],
            ["--FLASHONLY"],
            ["--SKIPPREBUILD"],
            ["--SKIPPOSTBUILD"]
        ]
        for argpart in args:
            results = parserObj.parse_args(argpart)
            builder.RetrievePlatformCommandLineOptions(results)

    def test_go_skip_building(self):
        builder = uefi_build.UefiBuilder()
        builder.SkipPostBuild = True
        builder.SkipBuild = True
        builder.SkipBuild = True
        manager = PluginManager()
        shell_environment.GetBuildVars().SetValue("EDK_TOOLS_PATH", self.WORKSPACE, "empty")
        helper = uefi_helper_plugin.HelperFunctions()
        ret = builder.Go(self.WORKSPACE, "", helper, manager)
        self.assertEqual(ret, 0)

    def test_build_wrapper(self):
        """Tests that a build wrapper can be used."""
        builder = uefi_build.UefiBuilder()

        # Post-build is not needed to test the build wrapper
        builder.SkipPostBuild = True

        # Some basic build variables need to be set to make it through
        # the build preamble to the point the wrapper gets called.
        shell_environment.GetBuildVars().SetValue("TARGET_ARCH",
                                                  "IA32",
                                                  "Set in build wrapper test")
        shell_environment.GetBuildVars().SetValue("EDK_TOOLS_PATH",
                                                  self.WORKSPACE,
                                                  "Set in build wrapper test")

        # "build_wrapper" -> The actual build_wrapper script
        # "test_file" -> An empty file written by build_wrapper
        build_wrapper_path = os.path.join(self.WORKSPACE, "build_wrapper")
        test_file_path = os.path.join(self.WORKSPACE, "test_file")

        # This script will write an empty file called "test_file" to the
        # temporary directory (workspace) to demonstrate that it ran successfully
        build_wrapper_file_content = """
            import os
            import sys

            test_file_dir = os.path.dirname(os.path.realpath(__file__))
            test_file_path = os.path.join(test_file_dir, "test_file")

            with open(test_file_path, 'w'):
                pass

            sys.exit(0)
            """

        build_wrapper_cmd = "python"
        build_wrapper_params = os.path.normpath(build_wrapper_path)

        TestUefiBuild.write_to_file(
            build_wrapper_path,
            cleandoc(build_wrapper_file_content))

        if GetHostInfo().os == "Linux":
            os.chmod(build_wrapper_path,
                     os.stat(build_wrapper_path).st_mode | stat.S_IEXEC)

        # This is the main point of this test. The wrapper file should be
        # executed instead of the build command. In real scenarios, the wrapper
        # script would subsequently call the build command.
        shell_environment.GetBuildVars().SetValue(
            "EDK_BUILD_CMD", build_wrapper_cmd, "Set in build wrapper test")
        shell_environment.GetBuildVars().SetValue(
            "EDK_BUILD_PARAMS", build_wrapper_params, "Set in build wrapper test")

        manager = PluginManager()
        helper = uefi_helper_plugin.HelperFunctions()
        ret = builder.Go(self.WORKSPACE, "", helper, manager)

        # Check the build wrapper return code
        self.assertEqual(ret, 0)

        # Check that the build wrapper ran successfully by checking that the
        # file written by the build wrapper file exists
        self.assertTrue(os.path.isfile(test_file_path))

    # TODO finish unit test


if __name__ == '__main__':
    unittest.main()



# @file test_signtool_signer.py
# This contains unit tests for the signtool binary wrapper
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import os
import sys
from edk2toolext.capsule import signtool_signer


class Test_signtool_signer(unittest.TestCase):

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_get_path(self):
        path = signtool_signer.get_signtool_path()
        self.assertTrue(os.path.exists(path))

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_sign_with_bad_options(self):
        signature = {
            "type": "test"
        }
        signer = {}
        with self.assertRaises(ValueError):
            signtool_signer.sign(None, signature, signer)

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_sign_with_good_options(self):
        signature = {
            "type": "pkcs7",
            "type_options": ["embedded"],
            "encoding": "DER",
            "hash_alg": "sha256"
        }
        signer = {
            "key_file": "file.txt",
            "key_file_format": "pkcs12"
        }
        with self.assertRaises(RuntimeError):
            signtool_signer.sign(b"data", signature, signer)

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_sign_with_mutually_exclusive_options(self):
        signature = {
            "type": "pkcs7",
            "type_options": ["embedded", "detachedSignedData"]
        }
        signer = {}
        with self.assertRaises(ValueError):
            signtool_signer.sign(b"data", signature, signer)

        signature = {
            "type": "pkcs7",
            "type_options": ["pkcs7DetachedSignedData", "detachedSignedData"]
        }
        signer = {}
        with self.assertRaises(ValueError):
            signtool_signer.sign(b"data", signature, signer)

        signature = {
            "type": "pkcs7",
            "type_options": ["pkcs7DetachedSignedData", "embedded"]
        }
        signer = {}
        with self.assertRaises(ValueError):
            signtool_signer.sign(b"data", signature, signer)

        signature = {
            "type": "pkcs7",
            "type_options": ["detachedSignedData", "pkcs7DetachedSignedData", "embedded"]
        }
        signer = {}
        with self.assertRaises(ValueError):
            signtool_signer.sign(b"data", signature, signer)



# @file test_ci_build_plugin.py
# Unit test for the ci build plugin class
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import os
import tempfile
import shutil
from edk2toolext.environment.plugintypes.ci_build_plugin import ICiBuildPlugin


class TestICiBuildPlugin(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        self.test_dir = None
        super().__init__(*args, **kwargs)

    def prep_workspace(self):
        self.clean_workspace()
        self.test_dir = tempfile.mkdtemp()

    def clean_workspace(self):
        if self.test_dir is None:
            return
        if os.path.isdir(self.test_dir):
            shutil.rmtree(self.test_dir)
            self.test_dir = None

    def setUp(self):
        self.prep_workspace()

    def tearDown(self):
        self.clean_workspace()

    def test_basic_api(self):
        plugin = ICiBuildPlugin()
        self.assertIsNone(plugin.GetTestName("", ""))
        self.assertIsNone(plugin.RunBuildPlugin("", "", "", "", "", "", "", ""))
        self.assertIn("NO-TARGET", plugin.RunsOnTargetList())

    def test_invalid_parameters_WalkDirectoryForExtension(self):
        plugin = ICiBuildPlugin()

        with self.assertRaises(TypeError) as context:
            plugin.WalkDirectoryForExtension("", "", "")
        self.assertTrue("extensionlist must be a list" in str(context.exception))

        with self.assertRaises(TypeError) as context:
            plugin.WalkDirectoryForExtension(["test"], None, "")
        self.assertTrue("directory is None" in str(context.exception))

        with self.assertRaises(ValueError) as context:
            plugin.WalkDirectoryForExtension(["test"], "a/b/c", "")
        self.assertTrue("directory is not an absolute path" in str(context.exception))

        with self.assertRaises(ValueError) as context:
            plugin.WalkDirectoryForExtension(["test"], os.path.join(self.test_dir, "junkdir", "junk"), "")
        self.assertTrue("directory is not a valid directory path" in str(context.exception))

        with self.assertRaises(TypeError) as context:
            plugin.WalkDirectoryForExtension([".py"], self.test_dir, "")
        self.assertTrue("ignorelist must be a list" in str(context.exception))

    def test_valid_parameters_WalkDirectoryForExtension(self):
        plugin = ICiBuildPlugin()

        with open(os.path.join(self.test_dir, "junk.txt"), "w") as the_file:
            the_file.write("Hello")

        nestedfolder = os.path.join(self.test_dir, "the_dir")
        os.makedirs(nestedfolder)
        with open(os.path.join(nestedfolder, "file2.py"), "w") as the_file:
            the_file.write("hello 2")

        # no files of this filetype found
        result = plugin.WalkDirectoryForExtension([".c"], self.test_dir)
        self.assertEqual(len(result), 0)

        # 1 txt file found
        result = plugin.WalkDirectoryForExtension([".txt"], self.test_dir)
        self.assertEqual(len(result), 1)

        # 1 txt and 1 py file found
        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, [])
        self.assertEqual(len(result), 2)

        # case insensitive file extension
        result = plugin.WalkDirectoryForExtension([".tXt", ".PY"], self.test_dir)
        self.assertEqual(len(result), 2)

    def test_valid_parameters_ignore_WalkDirectoryForExtension(self):
        plugin = ICiBuildPlugin()

        # setup test dir with a few files
        with open(os.path.join(self.test_dir, "junk.txt"), "w") as the_file:
            the_file.write("Hello")

        nestedfolder = os.path.join(self.test_dir, "the_dir")
        os.makedirs(nestedfolder)
        with open(os.path.join(nestedfolder, "file2.py"), "w") as the_file:
            the_file.write("hello 2")

        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, ["junk"])
        self.assertEqual(len(result), 1)

        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, ["file2"])
        self.assertEqual(len(result), 1)

        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, ["junk", "file2"])
        self.assertEqual(len(result), 0)

    def test_valid_parameters_ignore_caseinsensitive_WalkDirectoryForExtension(self):
        plugin = ICiBuildPlugin()

        # setup test dir with a few files
        with open(os.path.join(self.test_dir, "junk.txt"), "w") as the_file:
            the_file.write("Hello")

        nestedfolder = os.path.join(self.test_dir, "the_dir")
        os.makedirs(nestedfolder)
        with open(os.path.join(nestedfolder, "file2.py"), "w") as the_file:
            the_file.write("hello 2")

        # case insensitive
        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, ["JUNK"])
        self.assertEqual(len(result), 1)

        # case insensitive + partial match
        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, ["FILE"])
        self.assertEqual(len(result), 1)

        # case insensitive + all match including extension
        result = plugin.WalkDirectoryForExtension([".txt", ".py"], self.test_dir, ["FILE2.py"])
        self.assertEqual(len(result), 1)



# @file test_nuget_publish.py
# Unit test suite for the NugetSupport class.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
from edk2toolext import nuget_publishing
import sys
import tempfile
import os


class test_nuget_publish(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    @classmethod
    def write_to_file(cls, path, contents, close=True):
        f = open(path, "w")
        f.writelines(contents)
        if close:
            f.close()

    def test_init(self):
        nuget = nuget_publishing.NugetSupport("test")
        self.assertIsNotNone(nuget)
        # make sure we raise if we don't pass in anything
        with self.assertRaises(ValueError):
            nuget_publishing.NugetSupport()
        # write the config to file so we can read it into a new nuget support
        _, tempfile_path = tempfile.mkstemp(text=True)
        nuget.ConfigChanged = True
        nuget.ToConfigFile(tempfile_path)
        # read in a config file
        nuget2 = nuget_publishing.NugetSupport(ConfigFile=tempfile_path)
        self.assertIsNotNone(nuget2)

    def test_go_new(self):
        args = sys.argv
        sys.argv = [""]
        try:
            nuget_publishing.go()
        except SystemExit:
            # we'll fail because we don't pass in any arguments
            pass
        sys.argv = args

    def test_print(self):
        nuget = nuget_publishing.NugetSupport("test")
        nuget.Print()

    def test_empty_pack(self):
        nuget = nuget_publishing.NugetSupport("test")
        version = "1.1"
        nuget.SetBasicData("EDK2", "https://BSD2", "https://project_url", "descr", "server", "copyright")
        tempfolder_in = tempfile.mkdtemp()
        tempfolder_out = tempfile.mkdtemp()
        # this should fail because we don't have anything to pack
        ret = nuget.Pack(version, tempfolder_out, tempfolder_in)
        self.assertEqual(ret, 1)

    def test_change_copyright(self):
        nuget = nuget_publishing.NugetSupport("test")
        nuget.UpdateCopyright("ALL RIGHTS RESERVED.")
        self.assertTrue(nuget.ConfigChanged)

    def test_change_tags(self):
        nuget = nuget_publishing.NugetSupport("test")
        nuget.UpdateTags(["TAG1", "TAG2"])
        self.assertTrue(nuget.ConfigChanged)

    def test_push_without_spec(self):
        nuget = nuget_publishing.NugetSupport("test")
        tempfolder_out = tempfile.mkdtemp()
        spec = os.path.join(tempfolder_out, "test.nuspec")
        with self.assertRaises(Exception):
            ret = nuget.Push(spec, "")
            self.assertEqual(ret, 1)

    def test_push(self):
        nuget = nuget_publishing.NugetSupport("test")
        nuget.SetBasicData("EDK2", "https://BSD2", "https://project_url", "descr", "https://server", "copyright")
        tempfolder_out = tempfile.mkdtemp()
        spec = os.path.join(tempfolder_out, "test.nuspec")
        test_nuget_publish.write_to_file(spec, ["This is a legit nuget file lol", ])
        ret = nuget.Push(spec, "")
        self.assertEqual(ret, 1)

    def test_pack(self):
        nuget = nuget_publishing.NugetSupport("test")
        version = "1.1.1"
        nuget.SetBasicData("EDK2", "https://BSD2", "https://project_url", "description", "server", "copyright")
        tempfolder_in = tempfile.mkdtemp()
        tempfolder_out = tempfile.mkdtemp()
        outfile = os.path.join(tempfolder_in, "readme.txt")
        # Create a very long release notes
        release_notes = ""
        while len(release_notes) <= nuget_publishing.NugetSupport.RELEASE_NOTE_SHORT_STRING_MAX_LENGTH:
            release_notes += f"This is now {len(release_notes)} characters long. "
        # write a file that can be packaged by nuget
        test_nuget_publish.write_to_file(outfile, [release_notes, ])
        ret = nuget.Pack(version, tempfolder_out, tempfolder_in, release_notes)
        self.assertEqual(ret, 0)
        spec = os.path.join(tempfolder_out, "test.nuspec")
        self.assertTrue(os.path.exists(spec))
        # test that we clean up the files we aren't using
        nuget.CleanUp()
        self.assertFalse(os.path.exists(spec))

    def test_main_new(self):
        args = sys.argv
        tempfolder = tempfile.mkdtemp()
        sys.argv = ["",
                    "--Operation",
                    "New",
                    "--Name",
                    "Test",
                    "--Author",
                    "test",
                    "--ProjectUrl",
                    "https://github.com",
                    "--Description",
                    "test",
                    "--FeedUrl",
                    " https://github.com",
                    "--ConfigFileFolderPath",
                    tempfolder,
                    "--LicenseType",
                    "BSD2"]
        ret = nuget_publishing.main()
        self.assertEqual(ret, 0)
        sys.argv = args

    # TODO: finish unit test


if __name__ == '__main__':
    unittest.main()



# @file repo_resolver.py
# This module supports git operations (git repos).
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This module supports git operations."""
import os
import logging
from edk2toolext import edk2_logging
from edk2toolext.edk2_git import Repo
from edk2toollib.utility_functions import RemoveTree

# this follows a documented flow chart


# checks out dependency at git_path
def resolve(file_system_path, dependency, force=False, ignore=False, update_ok=False):
    """Resolves a particular repo.

    Args:
        file_system_path (Pathlike): path to repo
        dependency (Dict): contains Path, Commit, Branch
        force (bool): If it is OK to update the commit or branch
        ignore (bool): If it is OK to ignore errors or not.
        update_ok (bool): If it is OK to update the commit or branch

    Raises:
        (Exception): An error resolving a repo and ignore=False
    """
    logger = logging.getLogger("git")
    logger.info("Checking for dependency {0}".format(dependency["Path"]))
    git_path = os.path.abspath(file_system_path)

    # check if we have a path in our dependency
    if "Path" in dependency and not git_path.endswith(os.path.relpath(dependency["Path"])):
        # if we don't already the the path from the dependency at the end of the path we've been giving
        git_path = os.path.join(git_path, dependency["Path"])
    logging.info(f"Resolving at {git_path}")
    ##
    # NOTE - this process is defined in the Readme.md including flow chart for this behavior
    ##
    if not os.path.isdir(git_path):
        logging.info(f"Cloning at {git_path}")
        _, r = clone_repo(git_path, dependency)
        checkout(git_path, dependency, r, True, False)
        return r

    folder_empty = len(os.listdir(git_path)) == 0
    if folder_empty:  # if the folder is empty, we can clone into it
        _, r = clone_repo(git_path, dependency)
        checkout(git_path, dependency, r, True, False)
        return r

    repo = Repo(git_path)
    if not repo.initalized:  # if there isn't a .git folder in there
        if force:
            clear_folder(git_path)
            logger.warning(
                "Folder {0} is not a git repo and is being overwritten!".format(git_path))
            _, r = clone_repo(git_path, dependency)
            checkout(git_path, dependency, repo, True, False)
            return repo
        else:
            if (ignore):
                logger.warning(
                    "Folder {0} is not a git repo but Force parameter not used.  "
                    "Ignore State Allowed.".format(git_path))
                return repo
            else:
                logger.critical(
                    "Folder {0} is not a git repo and it is not empty.".format(git_path))
                raise Exception(
                    "Folder {0} is not a git repo and it is not empty".format(git_path))

    if repo.dirty:
        if force:
            clear_folder(git_path)
            logger.warning(
                "Folder {0} is a git repo but is dirty and is being overwritten as requested!".format(git_path))
            _, r = clone_repo(git_path, dependency)
            checkout(git_path, dependency, repo, True, False)
            return repo
        else:
            if (ignore):
                logger.warning(
                    "Folder {0} is a git repo but is dirty and Force parameter not used.  "
                    "Ignore State Allowed.".format(git_path))
                return repo
            else:
                logger.critical(
                    "Folder {0} is a git repo and is dirty.".format(git_path))
                raise Exception(
                    "Folder {0} is a git repo and is dirty.".format(git_path))

    if repo.remotes.origin.url != dependency["Url"]:
        if force:
            clear_folder(git_path)
            logger.warning(
                "Folder {0} is a git repo but it is at a different repo and is "
                "being overwritten as requested!".format(git_path))
            clone_repo(git_path, dependency)
            checkout(git_path, dependency, repo, True, False)
        else:
            if ignore:
                logger.warning(
                    "Folder {0} is a git repo pointed at a different remote.  "
                    "Can't checkout or sync state".format(git_path))
                return
            else:
                logger.critical("The URL of the git Repo {2} in the folder {0} does not match {1}".format(
                    git_path, dependency["Url"], repo.remotes.origin.url))
                raise Exception("The URL of the git Repo {2} in the folder {0} does not match {1}".format(
                    git_path, dependency["Url"], repo.remotes.origin.url))
    # if we've gotten here, we should just checkout as normal
    checkout(git_path, dependency, repo, update_ok, ignore, force)
    return repo

##
# dependencies is a list of objects - it has Path, Commit, Branch,


def resolve_all(workspace_path, dependencies, force=False, ignore=False, update_ok=False, omnicache_dir=None):
    """Resolves all repos.

    Args:
        workspace_path (Pathlike): workspace root
        dependencies (List[Dict]): Dict contains Path, Commit, Branch
        force (bool): If it is OK to update the commit or branch
        ignore (bool): If it is OK to ignore errors or not.
        update_ok (bool): If it is OK to update the commit or branch
        omnicache_dir (:obj:`bool`, optional): Omnicache path, if used

    Raises:
        (Exception): An error resolving a repo and ignore=False
    """
    logger = logging.getLogger("git")
    repos = []
    if force:
        logger.info("Resolving dependencies by force")
    if update_ok:
        logger.info("Resolving dependencies with updates as needed")
    for dependency in dependencies:
        dep_path = dependency["Path"]
        logger.log(edk2_logging.PROGRESS, f"Syncing {dep_path}")
        if "ReferencePath" not in dependency and omnicache_dir:
            dependency["ReferencePath"] = omnicache_dir
        if "ReferencePath" in dependency:  # make sure that the omnicache dir is relative to the working directory
            dependency["ReferencePath"] = os.path.join(workspace_path, dependency["ReferencePath"])
        git_path = os.path.join(workspace_path, dep_path)
        repos.append(git_path)
        resolve(git_path, dependency, force, ignore, update_ok)

    # print out the details- this is optional
    for dependency in dependencies:
        git_path = os.path.join(workspace_path, dependency["Path"])
        GitDetails = get_details(git_path)
        # print out details
        logger.info("{3} = Git Details: Url: {0} Branch {1} Commit {2}".format(
            GitDetails["Url"], GitDetails["Branch"], GitDetails["Commit"], dependency["Path"]))

    return repos


def get_details(abs_file_system_path):
    """Gets the Url, Branch, and Commit of a particular repo.

    Args:
        abs_file_system_path (PathLike): repo directory

    Returns:
        (Dict): Url, Branch, Commit
    """
    repo = Repo(abs_file_system_path)
    url = repo.remotes.origin.url
    active_branch = repo.active_branch
    head = repo.head.commit
    return {"Url": url, "Branch": active_branch, "Commit": head}


def clear_folder(abs_file_system_path):
    """Cleans the folder.

    Args:
        abs_file_system_path (PathLike): Directory to delete.
    """
    logger = logging.getLogger("git")
    logger.warning("WARNING: Deleting contents of folder {0} to make way for Git repo".format(
        abs_file_system_path))
    RemoveTree(abs_file_system_path)


def clone_repo(abs_file_system_path, DepObj):
    """Clones the repo in the folder using the dependency object.

    Args:
        abs_file_system_path (PathLike): destination to clone
        DepObj (Dict): dict containing Commit, Full, Branch, etc

    Returns:
        (Tuple[PathLike, bool]): (destination, result)
    """
    logger = logging.getLogger("git")
    logger.log(edk2_logging.get_progress_level(), "Cloning repo: {0}".format(DepObj["Url"]))
    dest = abs_file_system_path
    if not os.path.isdir(dest):
        os.makedirs(dest, exist_ok=True)
    shallow = False
    branch = None
    if "Commit" in DepObj:
        shallow = False
    if "Full" in DepObj and DepObj["Full"] is True:
        shallow = False
    if "Branch" in DepObj:
        shallow = True
        branch = DepObj["Branch"]

    reference = None
    if "ReferencePath" in DepObj and os.path.exists(DepObj["ReferencePath"]):
        reference = os.path.abspath(DepObj["ReferencePath"])
    result = Repo.clone_from(DepObj["Url"], dest, branch=branch, shallow=shallow, reference=reference)

    if result is None:
        if "ReferencePath" in DepObj:
            # attempt a retry without the reference
            logger.warning("Reattempting to clone without a reference. {0}".format(DepObj["Url"]))
            result = Repo.clone_from(DepObj["Url"], dest, branch=branch, shallow=shallow)
            if result is None:
                return (dest, None)

    return (dest, result)


def checkout(abs_file_system_path, dep, repo, update_ok=False, ignore_dep_state_mismatch=False, force=False):
    """Checks out a commit or branch.

    Args:
        abs_file_system_path (PathLike): The path to the repo
        dep (Dict): A dictionary containing a either a Commit or Branch, and also a Path
        repo (Repo): A valid repo object
        update_ok (bool): If it is OK to update the commit or branch
        ignore_dep_state_mismatch (bool): Whether a mismatch will result in an exception or not.
        force (bool): If it is OK to update the commit or branch

    Raises:
        (Exception): dependency state mismatch if ignore_dep_state_mismatch = False

    TIP: Either abs_file_system_path or repo is necessary. Not both.
    """
    logger = logging.getLogger("git")
    if repo is None:
        repo = Repo(abs_file_system_path)
    if "Commit" in dep:
        commit = dep["Commit"]
        if update_ok or force:
            repo.fetch()
            result = repo.checkout(commit=commit)
            if result is False:
                repo.fetch()
                repo.checkout(commit=commit)
            repo.submodule("update", "--init", "--recursive")
        else:
            if repo.head.commit == commit:
                logger.debug(
                    "Dependency {0} state ok without update".format(dep["Path"]))
                return
            elif ignore_dep_state_mismatch:
                logger.warning(
                    "Dependency {0} is not in sync with requested commit.  Ignore state allowed".format(dep["Path"]))
                return
            else:
                logger.critical(
                    "Dependency {0} is not in sync with requested commit.  Fail.".format(dep["Path"]))
                raise Exception(
                    "Dependency {0} is not in sync with requested commit.  Fail.".format(dep["Path"]))

    elif "Branch" in dep:
        branch = dep["Branch"]
        if update_ok or force:
            repo.fetch()
            result = repo.checkout(branch=branch)
            if result is False:  # we failed to do this
                # try to fetch it and try to checkout again
                logger.info("We failed to checkout this branch, we'll try to fetch")
                repo.fetch(branch=branch)
                repo.checkout(branch=branch)
            repo.submodule("update", "--init", "--recursive")
        else:
            if repo.active_branch == dep["Branch"]:
                logger.debug(
                    "Dependency {0} state ok without update".format(dep["Path"]))
                return
            elif ignore_dep_state_mismatch:
                logger.warning(
                    "Dependency {0} is not in sync with requested branch.  Ignore state allowed".format(dep["Path"]))
                return
            else:
                error = "Dependency {0} is not in sync with requested branch. Expected: {1}. Got {2} Fail.".format(
                    dep["Path"], dep["Branch"], repo.active_branch)
                logger.critical(error)
                raise Exception(error)
    else:
        raise Exception(
            "Branch or Commit must be specified for {0}".format(dep["Path"]))



# @file test_image_validation.py
# This contains unit tests for the image validation tool.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import edk2toolext.image_validation as IV


class Section:
    def __init__(self, name, characteristics=None):
        self.Characteristics = characteristics
        self.Name = name


class OptionalHeader:
    def __init__(self, SectionAlignment=None, Subsystem=None, DllCharacteristics=None):
        self.SectionAlignment = SectionAlignment
        self.Subsystem = Subsystem
        self.DllCharacteristics = DllCharacteristics


class FileHeader:
    def __init__(self):
        self.Machine = 0x8664

# A Dummy class to represent a PE.


class PE:
    def __init__(self, sections=None, optional_header=None):
        self.sections = sections
        self.OPTIONAL_HEADER = optional_header
        self.FILE_HEADER = FileHeader()

    def merge_modified_section_data(self):
        pass


class TestImageValidationInterface(unittest.TestCase):

    def test_add_test(self):

        test_manager = IV.TestManager()
        self.assertEqual(len(test_manager.tests), 0)
        test_manager.add_test(IV.TestSectionAlignment())
        self.assertEqual(len(test_manager.tests), 1)

    def test_add_tests(self):
        test_manager = IV.TestManager()
        self.assertEqual(len(test_manager.tests), 0)
        test_manager.add_tests([IV.TestSectionAlignment(), IV.TestSubsystemValue()])
        self.assertEqual(len(test_manager.tests), 2)

    def test_test_manager(self):
        config_data = {
            "TARGET_ARCH": {
                "X64": "IMAGE_FILE_MACHINE_AMD64"
            },
            "IMAGE_FILE_MACHINE_AMD64": {
                "DEFAULT": {
                    "DATA_CODE_SEPARATION": False
                }
            }
        }
        test_manager = IV.TestManager(config_data=config_data)
        self.assertEqual(test_manager.config_data, config_data)

        pe = PE(sections=[Section("S1.1".encode("utf-8"), characteristics=0x80000000)])
        test_manager.add_test(IV.TestWriteExecuteFlags())
        result = test_manager.run_tests(pe, "BAD_PROFILE")
        self.assertEqual(result, IV.Result.FAIL)

        result = test_manager.run_tests(pe)
        self.assertEqual(result, IV.Result.PASS)

    def test_write_execute_flags_test(self):
        """
        TestWriteExecuteFlags follows the following logic:
        1. If test requirement is not specified, or equal to false, return Result.SKIP
        3. Return Result.PASS / Result.FAIL returned based on Characteristic value
        """
        test_pe0 = PE(sections=[
            Section("S1.1".encode("utf-8"), characteristics=0x80000000),
            Section("S1.2".encode("utf-8"), characteristics=0x20000000),
            Section("S1.3".encode("utf-8"), characteristics=0x00000000)])

        test_pe1 = PE(sections=[
            Section("S2.1".encode("utf-8"), characteristics=0xA0000000),
            Section("S2.2".encode("utf-8"), characteristics=0x20000000),
            Section("S2.3".encode("utf-8"), characteristics=0x00000000)])

        test_pe2 = PE(sections=[
            Section("S3.1".encode("utf-8"), characteristics=0x20000000),
            Section("S3.2".encode("utf-8"), characteristics=0x80000000),
            Section("S3.3".encode("utf-8"), characteristics=0xC0000000)])

        test_pe3 = PE(sections=[
            Section("S4.1".encode("utf-8"), characteristics=0xE0000000)])

        config_data1 = {
            "TARGET_REQUIREMENTS": {"DATA_CODE_SEPARATION": False}
        }

        config_data2 = {"TARGET_REQUIREMENTS": {}}

        test_write_execute_flags = IV.TestWriteExecuteFlags()
        tests = [(test_pe0, IV.Result.PASS), (test_pe1, IV.Result.FAIL),
                 (test_pe2, IV.Result.PASS), (test_pe3, IV.Result.FAIL)]

        config_data0 = {
            "TARGET_REQUIREMENTS": {"DATA_CODE_SEPARATION": True}
        }

        # Test set 1
        for i in range(len(tests)):
            with self.subTest("test_write_execute1", i=i):
                pe, result = tests[i]
                self.assertEqual(test_write_execute_flags.execute(pe, config_data0), result)

        # Test set 2
        for i in range(len(tests)):
            with self.subTest("test_write_execute2", i=i):
                pe, _ = tests[i]
                self.assertEqual(test_write_execute_flags.execute(pe, config_data1), IV.Result.SKIP)

        # Test set 3
        for i in range(len(tests)):
            with self.subTest("test_write_execute3", i=i):
                pe, _ = tests[i]
                self.assertEqual(test_write_execute_flags.execute(pe, config_data2), IV.Result.SKIP)

    def test_section_alignment_test(self):
        """
        TestSectionAlignment follows the following logic:
        1. If requirements are not specified, or is empty, return Result.SKIP
        2. If SectionAlignment is not present, or is 0, return Result.WARN
        3. Return Result.PASS / Result.FAIL returned based on alignment requirements
        """

        config_data0 = {
            "TARGET_REQUIREMENTS": {},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }
        config_data1 = {
            "TARGET_REQUIREMENTS": {"ALIGNMENT": []},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }
        config_data2 = {
            "TARGET_REQUIREMENTS": {"ALIGNMENT": [{"COMPARISON": ">=", "VALUE": 0}]},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        config_data3 = {
            "TARGET_REQUIREMENTS": {"ALIGNMENT": [{"COMPARISON": "==", "VALUE": 1}]},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        config_data4 = {
            "TARGET_REQUIREMENTS": {"ALIGNMENT": [
                {"COMPARISON": ">=", "VALUE": 0},
                {"COMPARISON": ">=", "VALUE": 64},
                {"COMPARISON": "<=", "VALUE": 8192}],
                "ALIGNMENT_LOGIC_SEP": "AND"},
            "TARGET_INFO": {}
        }

        config_data5 = {
            "TARGET_REQUIREMENTS": {"ALIGNMENT": [
                {"COMPARISON": ">=", "VALUE": 0},
                {"COMPARISON": ">=", "VALUE": 64},
                {"COMPARISON": "!=", "VALUE": 4096}],
                "ALIGNMENT_LOGIC_SEP": "AND"},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        test_pe0 = PE(optional_header=OptionalHeader(SectionAlignment=4096))
        test_pe1 = PE(optional_header=OptionalHeader(SectionAlignment=0))
        test_pe2 = PE(optional_header=None)

        test_section_alignment_test = IV.TestSectionAlignment()
        tests0 = [(config_data0, IV.Result.SKIP), (config_data1, IV.Result.SKIP),
                  (config_data2, IV.Result.PASS), (config_data3, IV.Result.FAIL),
                  (config_data4, IV.Result.PASS), (config_data5, IV.Result.FAIL)]

        # Test set 1
        for i in range(len(tests0)):
            with self.subTest("test_section_alignment_test0", i=i):
                config, result = tests0[i]
                self.assertEqual(test_section_alignment_test.execute(test_pe0, config), result)

        tests1 = [(config_data0, IV.Result.SKIP), (config_data1, IV.Result.SKIP),
                  (config_data2, IV.Result.WARN), (config_data3, IV.Result.WARN),
                  (config_data4, IV.Result.WARN), (config_data5, IV.Result.WARN)]

        # Test set 2
        for i in range(len(tests1)):
            with self.subTest("test_section_alignment_test1", i=i):
                config, result = tests1[i]
                self.assertEqual(test_section_alignment_test.execute(test_pe1, config), result)

        # Test set 3
        for i in range(len(tests1)):
            with self.subTest("test_section_alignment_test2", i=i):
                config, result = tests1[i]
                self.assertEqual(test_section_alignment_test.execute(test_pe2, config), result)

    def test_section_alignment_test2(self):

        target_config0 = {
            "TARGET_REQUIREMENTS": {
                "ALIGNMENT": [
                    {"COMPARISON": ">=", "VALUE": 0},
                    {"COMPARISON": "<=", "VALUE": 8192}],
                "ALIGNMENT_LOGIC_SEP": "AND"},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        target_config1 = {
            "TARGET_REQUIREMENTS": {
                "ALIGNMENT": [
                    {"COMPARISON": "==", "VALUE": 0},
                    {"COMPARISON": "<=", "VALUE": 8192}],
                "ALIGNMENT_LOGIC_SEP": "AND"},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        target_config2 = {
            "TARGET_REQUIREMENTS": {
                "ALIGNMENT": [
                    {"COMPARISON": "==", "VALUE": 32},
                    {"COMPARISON": "==", "VALUE": 4096}],
                "ALIGNMENT_LOGIC_SEP": "OR"},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        target_config3 = {
            "TARGET_REQUIREMENTS": {
                "ALIGNMENT": [
                    {"COMPARISON": "==", "VALUE": 31},
                    {"COMPARISON": "==", "VALUE": 61}],
                "ALIGNMENT_LOGIC_SEP": "OR"},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        target_config4 = {
            "TARGET_REQUIREMENTS": {
                "ALIGNMENT": [
                    {"COMPARISON": "==", "VALUE": 32},
                    {"COMPARISON": "==", "VALUE": 64}]},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        target_config5 = {
            "TARGET_REQUIREMENTS": {
                "ALIGNMENT": [
                    {"COMPARISON": "==", "VALUE": 32},
                    {"COMPARISON": "==", "VALUE": 64}],
                "ALIGNMENT_LOGIC_SEP": "AR"},
            "TARGET_INFO": {"MACHINE_TYPE": "", "PROFILE": {}}
        }

        test_section_alignment_test = IV.TestSectionAlignment()
        pe = PE(optional_header=OptionalHeader(SectionAlignment=4096))
        tests = [(target_config0, IV.Result.PASS), (target_config1, IV.Result.FAIL), (target_config2, IV.Result.PASS),
                 (target_config3, IV.Result.FAIL), (target_config4, IV.Result.FAIL), (target_config5, IV.Result.FAIL)]

        for i in range(len(tests)):
            with self.subTest("test_section_alignment_and_or_logic", i=i):
                config, result = tests[i]
                self.assertEqual(test_section_alignment_test.execute(pe, config), result)

    def test_subsystem_value_test(self):
        """
        TestSubsystemValue follows the following logic:
        1. If Allowed Subsystems are not specified, or empty, return Result.SKIP
        2. If subsystem type is not present, return Result.WARN
        3. If subsystem type is invalid, return Result.FAIL
        3. return Result.PASS / Result.FAIL returned based on subsystem value
        """
        config_data0 = {
            "TARGET_REQUIREMENTS": {}
        }
        config_data1 = {
            "TARGET_REQUIREMENTS": {"ALLOWED_SUBSYSTEMS": []}
        }
        config_data2 = {
            "TARGET_REQUIREMENTS": {"ALLOWED_SUBSYSTEMS": ["IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER"]}
        }
        config_data3 = {
            "TARGET_REQUIREMENTS": {"ALLOWED_SUBSYSTEMS":
                                    ["IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                                     "IMAGE_SUBSYSTEM_EFI_APPLICATION"]}
        }

        test_subsystem_value_test = IV.TestSubsystemValue()

        TEST_PE0 = PE(optional_header=OptionalHeader(Subsystem=10))
        tests0 = [(config_data0, IV.Result.SKIP), (config_data1, IV.Result.SKIP),
                  (config_data2, IV.Result.FAIL), (config_data3, IV.Result.PASS)]

        for i in range(len(tests0)):
            with self.subTest("test_subsystem_value0", i=i):
                config, result = tests0[i]
                self.assertEqual(test_subsystem_value_test.execute(TEST_PE0, config), result)

        TEST_PE1 = PE(optional_header=OptionalHeader(Subsystem="UEFI_"))
        tests1 = [(config_data0, IV.Result.SKIP), (config_data1, IV.Result.SKIP),
                  (config_data2, IV.Result.FAIL), (config_data3, IV.Result.FAIL)]

        for i in range(len(tests1)):
            with self.subTest("test_subsystem_value1", i=i):
                config, result = tests1[i]
                self.assertEqual(test_subsystem_value_test.execute(TEST_PE1, config), result)

        TEST_PE2 = PE(optional_header=OptionalHeader(Subsystem=None))
        tests2 = [(config_data0, IV.Result.SKIP), (config_data1, IV.Result.SKIP),
                  (config_data2, IV.Result.WARN), (config_data3, IV.Result.WARN)]

        for i in range(len(tests2)):
            with self.subTest("test_subsystem_value2", i=i):
                config, result = tests2[i]
                self.assertEqual(test_subsystem_value_test.execute(TEST_PE2, config), result)

    def test_helper_functions(self):

        data = 0b00000000
        results = [1, 2, 4, 8, 16, 32, 64]

        for i in range(len(results)):
            with self.subTest("test_set_bit1", i=i):
                self.assertEqual(IV.set_bit(data, i), results[i])

        data = 0b11111111
        result = 255
        for i in range(7):
            with self.subTest("test_set_bit2", i=i):
                self.assertEqual(IV.set_bit(data, i), result)

        data = 0b11111111
        result = [254, 253, 251, 247, 239, 223, 191, 127]

        for i in range(len(result)):
            with self.subTest("test_clear_bit1", i=i):
                self.assertEqual(IV.clear_bit(data, i), result[i])

        data = 0b00000000
        result = 0

        for i in range(7):
            with self.subTest("test_clear_bit2", i=i):
                self.assertEqual(IV.clear_bit(data, i), result)

        data = 0b11001101
        masks = [0b10100000, 0b10000001, 0b10101010, 0b11110000, 0b00000001]
        result = [False, True, False, False, True]

        for i in range(len(result)):
            with self.subTest("test_has_characteristic", i=i):
                self.assertEqual(IV.has_characteristic(data, masks[i]), result[i])

        default_config = {
            "T1": "T",
            "T2": "T",
            "T3": "T",
        }
        target_config = {
            "T2": "F"
        }
        final_config = {
            "T1": "T",
            "T2": "F",
            "T3": "T"
        }

        self.assertEqual(IV.fill_missing_requirements(default_config, target_config), final_config)

    def test_test_interface(self):
        c = IV.TestInterface()

        with self.assertRaises(NotImplementedError):
            c.name()
        with self.assertRaises(NotImplementedError):
            c.execute(1, 2)

    def test_get_cli_args(self):

        test1 = ["-i", "file.efi"]
        test2 = ["-i", "file.efi", "-d"]
        test3 = ["-i", "file.efi", "-p", "APP"]
        test4 = ["--file", "file.efi", "--set-nx-compat"]
        test5 = ["--file", "file.efi", "--get-nx-compat"]
        test6 = ["--file", "file.efi", "--clear-nx-compat"]

        args = IV.get_cli_args(test1)
        self.assertEqual(args.file, "file.efi")
        self.assertEqual(args.debug, False)
        self.assertEqual(args.profile, None)

        args = IV.get_cli_args(test2)
        self.assertEqual(args.file, "file.efi")
        self.assertEqual(args.debug, True)
        self.assertEqual(args.profile, None)

        args = IV.get_cli_args(test3)
        self.assertEqual(args.file, "file.efi")
        self.assertEqual(args.debug, False)
        self.assertEqual(args.profile, "APP")

        args = IV.get_cli_args(test4)
        self.assertEqual(args.file, "file.efi")
        self.assertEqual(args.set_nx_compat, True)
        self.assertEqual(args.get_nx_compat, False)
        self.assertEqual(args.clear_nx_compat, False)

        args = IV.get_cli_args(test5)
        self.assertEqual(args.file, "file.efi")
        self.assertEqual(args.set_nx_compat, False)
        self.assertEqual(args.get_nx_compat, True)
        self.assertEqual(args.clear_nx_compat, False)

        args = IV.get_cli_args(test6)
        self.assertEqual(args.file, "file.efi")
        self.assertEqual(args.set_nx_compat, False)
        self.assertEqual(args.get_nx_compat, False)
        self.assertEqual(args.clear_nx_compat, True)

    def test_nx_flag_commands(self):
        pe = PE(optional_header=OptionalHeader(DllCharacteristics=0x0100))
        pe = IV.clear_nx_compat_flag(pe)
        self.assertEqual(pe.OPTIONAL_HEADER.DllCharacteristics, 0x0)
        self.assertEqual(IV.get_nx_compat_flag(pe), 0)

        pe = PE(optional_header=OptionalHeader(DllCharacteristics=0x0000))
        pe = IV.set_nx_compat_flag(pe)
        self.assertEqual(pe.OPTIONAL_HEADER.DllCharacteristics, 0x0100)
        self.assertEqual(IV.get_nx_compat_flag(pe), 1)



## @file versioninfo_helper.py
# This module contains helper functions for creating VERSIONINFO resources
# from json files, along with the functions to output version information from
# PE/PE+ files.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

# spell-checker:ignore NOFONT, ntdef
"""This module contains functions for creating VERSIONINFO resources.

This module contains helper functions for creating VERSIONINFO resources
from json files, along with the functions to output version information from
PE/PE+ files.
"""
import pefile
import json
import logging
from datetime import datetime
from typing import Tuple


class PEStrings(object):
    """PE/PE+ field String literals.

    String literals for fields in PE/PE+ header and required fields for
    VERSIONINFO resource.
    """

    FILE_OS_STRINGS = {
        0x00010000: "VOS_DOS",
        0x00040000: "VOS_NT",
        0x00000001: "VOS__WINDOWS16",
        0x00000004: "VOS__WINDOWS32",
        0x00020000: "VOS_OS216",
        0x00030000: "VOS_OS232",
        0x00000002: "VOS__PM16",
        0x00000003: "VOS__PM32",
        0x00000000: "VOS_UNKNOWN",
        0x00010001: "VOS_DOS_WINDOWS16",
        0x00010004: "VOS_DOS_WINDOWS32",
        0x00040004: "VOS_NT_WINDOWS32",
        0x00020002: "VOS_OS216_PM16",
        0x00030003: "VOS_OS232_PM32"
    }

    FILE_TYPE_STRINGS = {
        0x00000001: "VFT_APP",
        0x00000002: "VFT_DLL",
        0x00000003: "VFT_DRV",
        0x00000004: "VFT_FONT",
        0x00000007: "VFT_STATIC_LIB",
        0x00000000: "VFT_UNKNOWN",
        0x00000005: "VFT_VXD"
    }

    FILE_SUBTYPE_NOFONT_STRINGS = {
        0x0000000A: "VFT2_DRV_COMM",
        0x00000004: "VFT2_DRV_DISPLAY",
        0x00000008: "VFT2_DRV_INSTALLABLE",
        0x00000002: "VFT2_DRV_KEYBOARD",
        0x00000003: "VFT2_DRV_LANGUAGE",
        0x00000005: "VFT2_DRV_MOUSE",
        0x00000006: "VFT2_DRV_NETWORK",
        0x00000001: "VFT2_DRV_PRINTER",
        0x00000009: "VFT2_DRV_SOUND",
        0x00000007: "VFT2_DRV_SYSTEM",
        0x0000000C: "VFT2_DRV_VERSIONED_PRINTER",
        0x00000000: "VFT2_UNKNOWN"
    }

    FILE_SUBTYPE_FONT_STRINGS = {
        0x00000001: "VFT2_FONT_RASTER",
        0x00000003: "VFT2_FONT_TRUETYPE",
        0x00000002: "VFT2_FONT_VECTOR",
        0x00000000: "VFT2_UNKNOWN"
    }

    VALID_SIGNATURE = 0xfeef04bd
    DEFAULT_TRANSLATION = "0x0409,0x04b0"
    DEFAULT_BLOCK_HEADER = "040904b0"

    # PE/PE+ field names
    SIGNATURE_STR = "SIGNATURE"
    STRUC_VERSION_STR = "STRUCVERSION"
    FILE_VERSION_MS_STR = "FILEVERSIONMS"
    FILE_VERSION_LS_STR = "FILEVERSIONLS"
    FILE_VERSION_STR = "FILEVERSION"
    PRODUCT_VERSION_MS_STR = "PRODUCTVERSIONMS"
    PRODUCT_VERSION_LS_STR = "PRODUCTVERSIONLS"
    PRODUCT_VERSION_STR = "PRODUCTVERSION"
    FILE_FLAG_MASK_STR = "FILEFLAGSMASK"
    FILE_FLAGS_STR = "FILEFLAGS"
    FILE_TYPE_STR = "FILETYPE"
    FILE_SUBTYPE_STR = "FILESUBTYPE"
    FILE_OS_STR = "FILEOS"
    FILE_DATE_MS_STR = "FILEDATEMS"
    FILE_DATE_LS_STR = "FILEDATELS"
    FILE_DATE_STR = "FILEDATE"
    VFT_FONT_STR = "VFT_FONT"
    TRANSLATION_STR = "Translation"
    RSRC_STR = ".rsrc"
    STRING_FILE_INFO_STR = "StringFileInfo"
    VAR_FILE_INFO_STR = "VarFileInfo"

    FILE_VERSION_MS_PEFILE = "FileVersionMS"
    FILE_VERSION_LS_PEFILE = "FileVersionLS"
    FILE_VERSION_PEFILE = "FileVersion"
    PRODUCT_VERSION_MS_PEFILE = "ProductVersionMS"
    PRODUCT_VERSION_LS_PEFILE = "ProductVersionLS"
    PRODUCT_VERSION_PEFILE = "ProductVersion"
    FILE_DATE_MS_PEFILE = "FileDateMS"
    FILE_DATE_LS_PEFILE = "FileDateLS"
    FILE_SUBTYPE_PEFILE = "FileSubtype"

    # Key values for interacting with pefile objects
    PE_ENCODING = "utf-8"
    PE_STRUCT_STR = "Structure"
    PE_VALUE_STR = "Value"
    BEGIN_STR = "BEGIN"
    END_STR = "END"
    BLOCK_STR = "BLOCK"
    VALUE_STR = "VALUE"
    MINIMAL_MODE_STR = "MINIMAL"

    # Validation requirements
    VERSIONFILE_REQUIRED_FIELDS = {
        FILE_VERSION_STR,
        STRING_FILE_INFO_STR,
        VAR_FILE_INFO_STR
    }

    VERSIONFILE_ALLOWED_FIELDS = {
        FILE_VERSION_STR,
        PRODUCT_VERSION_STR,
        FILE_FLAG_MASK_STR,
        FILE_FLAGS_STR,
        FILE_OS_STR,
        FILE_TYPE_STR,
        FILE_SUBTYPE_STR,
        STRING_FILE_INFO_STR,
        VAR_FILE_INFO_STR,
        SIGNATURE_STR,
        STRUC_VERSION_STR,
        FILE_DATE_STR,
        FILE_DATE_MS_STR,
        FILE_DATE_LS_STR
    }

    COMPANY_NAME_STR = "CompanyName"
    PRODUCT_NAME_STR = "ProductName"
    INTERNAL_NAME_STR = "InternalName"
    ORIGINAL_FILENAME_STR = "OriginalFilename"

    STRING_FILE_INFO_REQUIRED_FIELDS = {
        COMPANY_NAME_STR,
        ORIGINAL_FILENAME_STR,
    }

    STRING_FILE_INFO_ALLOWED_FIELDS = {
        "Comments",
        "CompanyName",
        "FileDescription",
        "FileVersion",
        "InternalName",
        "LegalCopyright",
        "LegalTrademarks",
        "OriginalFilename",
        "PrivateBuild",
        "ProductName",
        "ProductVersion",
        "SpecialBuild"
    }

    VALID_FILE_OS_VALUES = {
        "VOS_UNKNOWN",
        "VOS_DOS",
        "VOS_NT",
        "VOS__WINDOWS16",
        "VOS__WINDOWS32",
        "VOS_DOS_WINDOWS16",
        "VOS_DOS_WINDOWS32",
        "VOS_NT_WINDOWS32"
    }

    VALID_FILE_TYPE_VALUES = {
        "VFT_APP",
        "VFT_DLL",
        "VFT_DRV",
        "VFT_FONT",
        "VFT_STATIC_LIB",
        "VFT_UNKNOWN",
        "VFT_VXD"
    }

    VALID_SUBTYPE_VFT_DRV = {
        "VFT2_UNKNOWN",
        "VFT2_DRV_COMM",
        "VFT2_DRV_PRINTER",
        "VFT2_DRV_KEYBOARD",
        "VFT2_DRV_LANGUAGE",
        "VFT2_DRV_DISPLAY",
        "VFT2_DRV_MOUSE",
        "VFT2_DRV_NETWORK",
        "VFT2_DRV_SYSTEM",
        "VFT2_DRV_INSTALLABLE",
        "VFT2_DRV_SOUND",
        "VFT2_DRV_VERSIONED_PRINTER"
    }

    VALID_SUBTYPE_VFT_FONT = {
        "VFT2_UNKNOWN",
        "VFT2_FONT_RASTER",
        "VFT2_FONT_VECTOR",
        "VFT2_FONT_TRUETYPE"
    }

    VALID_LANG_ID = {
        0x0401, 0x0402,
        0x0403, 0x0404,
        0x0405, 0x0406,
        0x0407, 0x0408,
        0x0409, 0x040A,
        0x040B, 0x040C,
        0x040D, 0x040E,
        0x040F, 0x0410,
        0x0411, 0x0412,
        0x0413, 0x0414,
        0x0415, 0x0416,
        0x0417, 0x0418,
        0x0419, 0x041A,
        0x041B, 0x041C,
        0x041D, 0x041E,
        0x041F, 0x0420,
        0x0421, 0x0804,
        0x0807, 0x0809,
        0x080A, 0x080C,
        0x0C0C, 0x100C,
        0x0816, 0x081A,
        0x0810, 0x0813,
        0x0814
    }

    VALID_CHARSET_ID = {
        0x0000, 0x03A4,
        0x03B5, 0x03B6,
        0x04B0, 0x04E2,
        0x04E3, 0x04E4,
        0x04E5, 0x04E6,
        0x04E7, 0x04E8
    }


def validate_version_number(version_str: str) -> bool:
    """Helper function to check if a version string format is valid or not.

    Args:
        version_str (str): the version

    Returns:
        (bool): if the version string is valid or not
    """
    if version_str.count('.') != 3 and version_str.count(',') != 3:
        logging.error("Invalid version string: " + version_str + ". Version must be in form "
                      + "\"INTEGER.INTEGER.INTEGER.INTEGER\".")
        return False

    split = None
    if version_str.count('.') == 3:
        split = version_str.split(".")
    else:
        split = version_str.split(",")

    for substr in split:
        try:
            if int(substr) > 65535:
                logging.error("Integer overflow in version string: " + version_str + ".")
                return False
        except ValueError:
            logging.error("Invalid version string: " + version_str + ". Version must be in form \""
                          + " INTEGER.INTEGER.INTEGER.INTEGER\".")
            return False

    return True


def version_str_to_int(version_str: str) -> Tuple[int, int]:
    """Converts a valid version string to a number.

    Given a valid version string, returns raw version number

    Args:
        version_str (str): version string

    Result:
        (Tuple[int, int]): (32 MS bits, 32 LS bits)
    """
    split = None
    if version_str.count('.') == 3:
        split = version_str.split(".")
    else:
        split = version_str.split(",")

    ms = int(split[1]) + (int(split[0]) << 16)
    ls = int(split[3]) + (int(split[2]) << 16)
    return (ms, ls)


def hex_to_version_str(val: int) -> str:
    """Converts hex version number to string.

    Args:
        val (int): value to convert

    Returns:
        (str): string represention
    """
    return str(((val & ~0) >> 16) & 0xffff) + "." + str(val & 0xffff)


class PEObject(object):
    """Class for parsing PE/PE+ files.

    Gives functionality for reading VS_VERSIONINFO metadata and .rsrc section.
    """
    _pe: pefile.PE = None

    def __init__(self, filepath: str) -> None:
        """Initializes PE parser.

        Args:
            filepath (str): filepath to PE/PE+ file to parse
        """
        try:
            self._pe = pefile.PE(filepath)
        except pefile.PEFormatError as e:
            logging.error("Error loading PE: " + str(e))

    def _populate_entry(self, key: str, val: int, dict: dict) -> None:
        """Formats and inserts VERSIONINFO fields into dictionary.

        Args:
            key (str): dict key for value
            val (int): value to store in dict
            dict (dict): dictionary to store key/value pair
        """
        if key.upper() == PEStrings.FILE_OS_STR:
            if val in PEStrings.FILE_OS_STRINGS:
                dict[key] = PEStrings.FILE_OS_STRINGS[val]
                return
        elif key.upper() == PEStrings.FILE_TYPE_STR:
            if val in PEStrings.FILE_TYPE_STRINGS:
                dict[key] = PEStrings.FILE_TYPE_STRINGS[val]
                return
        elif key.upper() == PEStrings.FILE_VERSION_MS_STR:
            dict[PEStrings.FILE_VERSION_PEFILE] = hex_to_version_str(val)
            return
        elif key.upper() == PEStrings.FILE_VERSION_LS_STR:
            dict[PEStrings.FILE_VERSION_PEFILE] += "." + hex_to_version_str(val)
            return
        elif key.upper() == PEStrings.PRODUCT_VERSION_MS_STR:
            dict[PEStrings.PRODUCT_VERSION_PEFILE] = hex_to_version_str(val)
            return
        elif key.upper() == PEStrings.PRODUCT_VERSION_LS_STR:
            dict[PEStrings.PRODUCT_VERSION_PEFILE] += "." + hex_to_version_str(val)
            return
        dict[key] = hex(val)

    def contains_rsrc(self) -> bool:
        """Checks if PE object contains .rsrc section.

        Returns:
            (bool): whether a .rsrc section exists in the PE
        """
        for section in self._pe.sections:
            if section.Name.decode(PEStrings.PE_ENCODING).replace("\x00", "") == PEStrings.RSRC_STR:
                return True
        return False

    def get_version_dict(self) -> dict:
        """Parses PE/PE+ loaded into PEObject.

        Returns:
            (dict): dict containing metadata from header and .rsrc section
        """
        if not self._pe:
            logging.fatal("Cannot parse, PE not loaded.")
            return None

        result = {}
        try:
            vs_fixedfileinfo_dict = self._pe.VS_FIXEDFILEINFO[0].dump_dict()
            for key in vs_fixedfileinfo_dict.keys():
                # Skip sections that have dependencies
                if key == PEStrings.PE_STRUCT_STR or \
                   key == PEStrings.FILE_SUBTYPE_PEFILE or \
                   key == PEStrings.FILE_VERSION_LS_PEFILE or \
                   key == PEStrings.PRODUCT_VERSION_LS_PEFILE or \
                   key == PEStrings.FILE_DATE_LS_PEFILE:
                    continue

                self._populate_entry(key, vs_fixedfileinfo_dict[key][PEStrings.PE_VALUE_STR], result)

            # Resolve dependent fields
            if PEStrings.FILE_VERSION_MS_PEFILE in vs_fixedfileinfo_dict.keys() and \
               PEStrings.FILE_VERSION_LS_PEFILE in vs_fixedfileinfo_dict.keys():
                self._populate_entry(PEStrings.FILE_VERSION_LS_PEFILE,
                                     vs_fixedfileinfo_dict[PEStrings.FILE_VERSION_LS_PEFILE][PEStrings.PE_VALUE_STR], result) # noqa

            if PEStrings.PRODUCT_VERSION_MS_PEFILE in vs_fixedfileinfo_dict.keys() and \
               PEStrings.PRODUCT_VERSION_LS_PEFILE in vs_fixedfileinfo_dict.keys():
                self._populate_entry(PEStrings.PRODUCT_VERSION_LS_PEFILE,
                                     vs_fixedfileinfo_dict[PEStrings.PRODUCT_VERSION_LS_PEFILE][PEStrings.PE_VALUE_STR], result) # noqa

            if PEStrings.FILE_DATE_MS_PEFILE in vs_fixedfileinfo_dict.keys() and \
               PEStrings.FILE_DATE_LS_PEFILE in vs_fixedfileinfo_dict.keys():
                self._populate_entry(PEStrings.FILE_DATE_LS_PEFILE,
                                     vs_fixedfileinfo_dict[PEStrings.FILE_DATE_LS_PEFILE][PEStrings.PE_VALUE_STR], result) # noqa

            if PEStrings.FILE_SUBTYPE_PEFILE in vs_fixedfileinfo_dict.keys():
                file_subtype = vs_fixedfileinfo_dict[PEStrings.FILE_SUBTYPE_PEFILE][PEStrings.PE_VALUE_STR]
                if PEStrings.FILE_TYPE_STR in result and result[PEStrings.FILE_TYPE_STR] == PEStrings.VFT_FONT_STR:
                    if file_subtype in PEStrings.FILE_SUBTYPE_FONT_STRINGS:
                        result[PEStrings.FILE_SUBTYPE_PEFILE] = PEStrings.FILE_SUBTYPE_FONT_STRINGS[PEStrings.file_subtype] # noqa
                    else:
                        result[PEStrings.FILE_SUBTYPE_PEFILE] = file_subtype
                else:
                    if file_subtype in PEStrings.FILE_SUBTYPE_NOFONT_STRINGS.keys():
                        result[PEStrings.FILE_SUBTYPE_PEFILE] = PEStrings.FILE_SUBTYPE_NOFONT_STRINGS[file_subtype]
                    else:
                        result[PEStrings.FILE_SUBTYPE_PEFILE] = file_subtype

        except AttributeError:
            logging.warning("Could not find VS_FIXEDFILEINFO.")

        if self.contains_rsrc():
            for fileinfo in self._pe.FileInfo:
                for entry in fileinfo:
                    if entry.Key.decode(PEStrings.PE_ENCODING).replace("\x00", "") == PEStrings.STRING_FILE_INFO_STR:
                        stringfileinfo_dict = {}
                        for strTable in entry.StringTable:
                            for item in strTable.entries.items():
                                stringfileinfo_dict[item[0].decode(PEStrings.PE_ENCODING)] = item[1].decode(PEStrings.PE_ENCODING) # noqa
                        result[PEStrings.STRING_FILE_INFO_STR] = stringfileinfo_dict
                    elif entry.Key.decode(PEStrings.PE_ENCODING).replace("\x00", "") == PEStrings.VAR_FILE_INFO_STR:
                        varfileinfo_dict = {}
                        for var in entry.Var:
                            for item in var.entry.items():
                                varfileinfo_dict[item[0].decode(PEStrings.PE_ENCODING)] = item[1]
                        result[PEStrings.VAR_FILE_INFO_STR] = varfileinfo_dict
        else:
            logging.warning("File does not contain .rsrc section.")

        return result


class VERSIONINFOGenerator(object):
    """A Generator for a VERSIONINFO rc file via a given JSON settings file.

    Given a JSON file, object creates generator for VERSIONINFO rc file for the given JSON.
    Provides validation of some basic requirements and generates VERSIONINFO.rc source to be used with rc.exe.
    If "Minimal" attribute is set to false, JSON defines a complete rc section.
    Here is an example of a complete rc section:
    {
        "Minimal": "False",
        "FileVersion": "1.0.0.0",
        "ProductVersion": "1.0.0.0",
        "FileFlagsMask": "VS_FFI_FILEFLAGSMASK",
        "FileFlags": "0",
        "FileOS": "VOS_NT",
        "FileType": "VFT_DRV",
        "FileSubtype": "VFT2_DRV_SYSTEM",
        "StringFileInfo": {
            "CompanyName": "Example Company",
            "OriginalFilename": "ExampleApp.efi",
            "FileVersion": "1.0.0.0",
        },
        "VarFileInfo": {
            "Translation": "0x0409 0x04b0"
        }
    }

    StringFileInfo and VarFileInfo can both have many more entries.
    More information available at https://docs.microsoft.com/en-us/windows/win32/menurc/versioninfo-resource.

    If the "Minimal" attribute is set to True or is not present, JSON must define a FileVersion, ProductName,
    and CompanyName, but nothing else. Here is an example of a minimal JSON representation:
    {
        "FileVersion": "1.0.0.0",
        "CompanyName": "Example Company",
        "OriginalFilename": "ExampleApp.efi"
    }

    or equivalently:
    {
        "Minimal": "True",
        "FileVersion": "1.0.0.0",
        "CompanyName": "Example Company",
        "OriginalFilename": "ExampleApp.efi"
    }
    """
    _minimal_required_fields = {
        PEStrings.FILE_VERSION_STR.upper(),
        PEStrings.COMPANY_NAME_STR.upper(),
        PEStrings.ORIGINAL_FILENAME_STR.upper()
    }

    _version_dict = None

    def __init__(self, filepath: str) -> None:
        """Initializes the VERSIONINFO generator.

        Initializes the VERSIONINFO generator using a JSON file containing
        VERSIONINFO information in the format shown in the class description.

        Args:
            filepath (str): filepath to the JSON file
        """
        with open(filepath, "r") as jsonFile:
            data = jsonFile.read()
            try:
                #
                # Convert all keys to UPPERCASE for consistant usage
                #
                self._version_dict = dict({k.upper(): v for k, v in json.loads(data).items()})
            except json.decoder.JSONDecodeError as e:
                logging.error("Invalid JSON format, " + str(e))

    def validate(self) -> bool:
        """Checks if the loaded JSON file represents a valid VERSIONINFO resource.

        Returns:
            (bool): valid or not
        """
        valid = True

        # First Pass: Check to see if all required fields are present
        required = {string.upper() for string in PEStrings.VERSIONFILE_REQUIRED_FIELDS}
        allowed = {string.upper() for string in PEStrings.VERSIONFILE_ALLOWED_FIELDS}
        for key in self._version_dict.keys():
            if key.upper() not in allowed:
                logging.error("Invalid parameter: " + key + ".")
                valid = False
            else:
                if key.upper() in required:
                    required.remove(key.upper())

        for remaining in required:
            logging.error("Missing required parameter: " + remaining + ".")
            valid = False

        if PEStrings.STRING_FILE_INFO_STR.upper() in self._version_dict:
            required_string_fields = PEStrings.STRING_FILE_INFO_REQUIRED_FIELDS.copy()
            for key in self._version_dict[PEStrings.STRING_FILE_INFO_STR.upper()]:
                if key in required_string_fields:
                    required_string_fields.remove(key)

            for remaining in required_string_fields:
                logging.error("Missing required StringFileInfo parameter: " + remaining + ".")
                valid = False

        if not valid:
            return False

        # Second pass: Check to see if fields are valid
        valid = validate_version_number(self._version_dict[PEStrings.FILE_VERSION_STR])
        if PEStrings.PRODUCT_VERSION_STR in self._version_dict:
            valid = valid and validate_version_number(self._version_dict[PEStrings.PRODUCT_VERSION_STR])

        if PEStrings.FILE_OS_STR in self._version_dict:
            try:
                if int(self._version_dict[PEStrings.FILE_OS_STR], 0) not in PEStrings.FILE_OS_STRINGS.keys():
                    valid = False
                    logging.error("Invalid FILEOS value: " + self._version_dict[PEStrings.FILE_OS_STR] + ".")
            except ValueError:
                if self._version_dict[PEStrings.FILE_OS_STR] not in PEStrings.VALID_FILE_OS_VALUES:
                    valid = False
                    logging.error("Invalid FILEOS value: " + self._version_dict[PEStrings.FILE_OS_STR] + ".")

        if PEStrings.FILE_TYPE_STR in self._version_dict:
            try:
                if int(self._version_dict[PEStrings.FILE_TYPE_STR], 0) not in PEStrings.FILE_TYPE_STRINGS.keys():
                    valid = False
                    logging.error("Invalid FILETYPE value: " + self._version_dict[PEStrings.FILE_TYPE_STR] + ".")
            except ValueError:
                if self._version_dict[PEStrings.FILE_TYPE_STR] not in PEStrings.VALID_FILE_TYPE_VALUES:
                    valid = False
                    logging.error("Invalid FILETYPE value: " + self._version_dict[PEStrings.FILE_TYPE_STR] + ".")
            if PEStrings.FILE_SUBTYPE_STR in self._version_dict:
                if self._version_dict[PEStrings.FILE_TYPE_STR] == "VFT_DRV":
                    if self._version_dict[PEStrings.FILE_SUBTYPE_STR] not in PEStrings.VALID_SUBTYPE_VFT_DRV:
                        logging.error("Invalid FILESUBTYPE value for FILETYPE VFT_DRV: "
                                      + self._version_dict[PEStrings.FILE_SUBTYPE_STR] + ".")
                        valid = False
                elif self._version_dict[PEStrings.FILE_TYPE_STR] == "VFT_FONT":
                    if self._version_dict[PEStrings.FILE_SUBTYPE_STR] not in PEStrings.VALID_SUBTYPE_VFT_FONT:
                        logging.error("Invalid FILESUBTYPE value for FILETYPE VFT_FONT: "
                                      + self._version_dict[PEStrings.FILE_SUBTYPE_STR] + ".")
                        valid = False
                elif (self._version_dict[PEStrings.FILE_TYPE_STR] != "VFT_VXD"
                      and self._version_dict[PEStrings.FILE_SUBTYPE_STR] != 0):
                    logging.error("Invalid FILESUBTYPE value for FILETYPE "
                                  + self._version_dict[PEStrings.FILE_TYPE_STR] + ", value must be 0.")
                    valid = False
        elif PEStrings.FILE_SUBTYPE_STR in self._version_dict:
            logging.error("Missing parameter: must have FileType if FileSubtype defined.")
            valid = False

        if PEStrings.TRANSLATION_STR in self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()]:
            langid_set = self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()][PEStrings.TRANSLATION_STR].split(" ")
            try:
                if len(langid_set) != 2:
                    logging.error("Translation field must contain 2 space delimited hexidecimal bytes.")
                    valid = False
                elif (int(langid_set[0].replace('"', ''), 0) not in PEStrings.VALID_LANG_ID
                      or int(langid_set[1].replace('"', ''), 0) not in PEStrings.VALID_CHARSET_ID):
                    logging.error("Invalid language code: "
                                  + self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()][PEStrings.TRANSLATION_STR] + ".") # noqa
                    valid = False
            except ValueError:
                logging.error("Invalid language code: "
                              + self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()][PEStrings.TRANSLATION_STR] + ".") # noqa
                valid = False
        else:
            logging.error("Missing required parameter in VarFileInfo: Translation.")
            valid = False

        for field in self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()].keys():
            if field != PEStrings.TRANSLATION_STR:
                logging.error("Invalid VarFileInfo parameter: %s.", field)
                valid = False

        return valid

    def validate_minimal(self) -> bool:
        """Checks if the loaded JSON file represents a valid minimal VERSIONINFO resource.

        Returns:
            (bool): valid or not
        """
        valid = True
        required = self._minimal_required_fields.copy()
        for key in self._version_dict:
            if key not in required:
                valid = False
                logging.error("Invalid minimal parameter: %s.", key)
            else:
                required.remove(key.upper())

        if not valid:
            return False

        return validate_version_number(self._version_dict[PEStrings.FILE_VERSION_STR])

    def write_minimal(self, path: str, version: str) -> bool:
        """Write a minimal VERSIONINFO resource file.

        Args:
            path (str): path to write
            version (str): version

        Returns:
            (bool): result of writing
        """
        if not self.validate_minimal():
            logging.error("Invalid input, aborted.")
            return False

        # Comment block for tool generation
        out_str = "/* Auto-generated VERSIONINFO resource file.\n"
        out_str += "   Generated %s \n" % datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        out_str += "   Tool Version: %s */\n\n" % version

        # Add required includes for rc compiler
        out_str += "#ifdef RC_INVOKED\n"

        # Header
        out_str += "VS_VERSION_INFO\tVERSIONINFO\n"
        out_str += PEStrings.FILE_VERSION_STR + "\t"
        version = self._version_dict[PEStrings.FILE_VERSION_STR.upper()].split(".")
        if len(version) != 4:
            version = self._version_dict[PEStrings.FILE_VERSION_STR.upper()].split(",")
        out_str += version[0] + ',' + version[1] + ',' + version[2] + ',' + version[3] + "\n"

        # StringFileInfo
        out_str += "\n" + PEStrings.BEGIN_STR + "\n\t" \
                   + PEStrings.BLOCK_STR + " \"" + PEStrings.STRING_FILE_INFO_STR \
                   + "\"\n\t" + PEStrings.BEGIN_STR + "\n" + "\t\t" \
                   + PEStrings.BLOCK_STR + " \"" + PEStrings.DEFAULT_BLOCK_HEADER \
                   + "\"\n\t\t" + PEStrings.BEGIN_STR + "\n" + "\t\t" + PEStrings.VALUE_STR \
                   + ' "' + PEStrings.COMPANY_NAME_STR + '",\t"' \
                   + self._version_dict[PEStrings.COMPANY_NAME_STR.upper()] + "\"\n" \
                   + "\t\t" + PEStrings.VALUE_STR + ' "' + PEStrings.ORIGINAL_FILENAME_STR + '",\t"' \
                   + self._version_dict[PEStrings.ORIGINAL_FILENAME_STR.upper()] + "\"\n" \
                   + "\t\t" + PEStrings.END_STR + "\n\t" + PEStrings.END_STR + "\n\n"

        # VarFileInfo
        out_str += "\t" + PEStrings.BLOCK_STR + " \"" + PEStrings.VAR_FILE_INFO_STR + '"\n\t' \
                   + PEStrings.BEGIN_STR + "\n" + "\t\t" + PEStrings.VALUE_STR + ' "' \
                   + PEStrings.TRANSLATION_STR + '",\t' + PEStrings.DEFAULT_TRANSLATION + "\n" \
                   + "\t" + PEStrings.END_STR + "\n" + PEStrings.END_STR + "\n#endif"

        with open(path, "w") as out:
            out.write(out_str)

        return True

    def write(self, path: str, version: str) -> bool:
        """Encodes the loaded JSON and writes it to VERSION.rc.

        Encodes the loaded JSON and writes it to VERSION.rc, a resource file compatible with rc.exe.

        Args:
            path (str): path to the directory that VERSIONINFO.rc will be written to.
            version (str): version

        Returns:
            (bool): result of encoding and writing
        """
        if not self._version_dict:
            logging.error("Invalid input, aborted.")
            return False

        self._version_dict = dict((key.upper(), val) for key, val in self._version_dict.items())

        minimal = True
        for key in self._version_dict:
            if key.upper() == PEStrings.MINIMAL_MODE_STR:
                if self._version_dict[key].upper() == "FALSE":
                    minimal = False
                elif self._version_dict[key].upper() != "TRUE":
                    logging.error("Invalid value for 'Minimal', must be boolean.")
                    logging.error("Invalid input, aborted.")
                    return False

                del self._version_dict[key]
                break

        if minimal:
            return self.write_minimal(path, version)

        if not self.validate():
            logging.error("Invalid input, aborted.")
            return False

        # Comment block for tool generation
        out_str = "/* Auto-generated VERSIONINFO resource file.\n"
        out_str += "   Generated %s \n" % datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        out_str += "   Tool Version: %s */\n\n" % version

        # Add required includes for rc compiler
        out_str += "#include <ntdef.h>\n#include <winver.h>\n#ifdef RC_INVOKED\n"

        # Header fields
        out_str += "VS_VERSION_INFO\tVERSIONINFO\n"
        for param in self._version_dict.keys():
            if (param == PEStrings.STRING_FILE_INFO_STR.upper()
               or param == PEStrings.VAR_FILE_INFO_STR.upper()):
                continue
            if param == PEStrings.PRODUCT_VERSION_STR or param == PEStrings.FILE_VERSION_STR:
                out_str += param.upper() + "\t"
                version = self._version_dict[param].split(".")
                out_str += version[0] + ',' + version[1] + ',' + version[2] + ',' + version[3] + "\n"
            else:
                out_str += param.upper() + "\t" + str(self._version_dict[param]) + "\n"

        # StringFileInfo
        out_str += "\n" + PEStrings.BEGIN_STR + "\n\t"
        out_str += PEStrings.BLOCK_STR + " \"" + PEStrings.STRING_FILE_INFO_STR + "\"\n\t" + PEStrings.BEGIN_STR + "\n"

        language_code = ""
        for code in self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()][PEStrings.TRANSLATION_STR].split(" "):
            language_code += code.split("0x", 1)[1]

        out_str += "\t\t" + PEStrings.BLOCK_STR + " \"" + language_code + "\"\n\t\t" + PEStrings.BEGIN_STR + "\n"
        for field in self._version_dict[PEStrings.STRING_FILE_INFO_STR.upper()].keys():
            out_str += "\t\t" + PEStrings.VALUE_STR + " \"" + field + "\",\t\"" \
                       + self._version_dict[PEStrings.STRING_FILE_INFO_STR.upper()][field] + "\"\n"

        out_str += "\t\t" + PEStrings.END_STR + "\n\t" + PEStrings.END_STR + "\n\n"

        # VarFileInfo
        out_str += "\t" + PEStrings.BLOCK_STR
        out_str += " \"" + PEStrings.VAR_FILE_INFO_STR + "\"\n\t" + PEStrings.BEGIN_STR + "\n"
        language_tokens = self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()][PEStrings.TRANSLATION_STR].split(" ")
        for field in self._version_dict[PEStrings.VAR_FILE_INFO_STR.upper()].keys():
            out_str += "\t\t" + PEStrings.VALUE_STR + " \"" + field + "\",\t" + language_tokens[0] + "," \
                       + language_tokens[1] + "\n"

        out_str += "\t" + PEStrings.END_STR + "\n" + PEStrings.END_STR + "\n#endif"

        with open(path, "w") as out:
            out.write(out_str)

        return True
